/*
 * Class: PersistentBloatContext 
 * Version: 
 * Date:
 * 
 * All files in the distribution of BLOAT (Bytecode Level Optimization and
 * Analysis tool for Java(tm)) are Copyright 1997-2001 by the Purdue
 * Research Foundation of Purdue University.  All rights reserved.
 *
 * <p>
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this entire copyright notice is duplicated in all
 * such copies, and that any documentation, announcements, and other
 * materials related to such distribution and use acknowledge that the
 * software was developed at Purdue University, West Lafayette, IN by
 * @author Antony Hosking
 * @author David Whitlock
 * @author Nathaniel Nystrom
 * No charge may be made for copies, derivations, or distributions of
 * this material without the express written consent of the copyright
 * holder.  Neither the name of the University nor the name of the
 * author may be used to endorse or promote products derived from this
 * material without specific prior written permission.  THIS SOFTWARE
 * IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR ANY PARTICULAR PURPOSE.
 *
 * <p>
 * 
 * Port of the BLOAT to use the Jakata Apache project BCEL 5.2 was 
 * contributed by The Australian National University by
 * @author Arrin Daley
 * @author John N Zigman
 * 
 * <p>
 * Java is a trademark of Sun Microsystems, Inc.
 */

package edu.purdue.cs.bloat.context;

import org.apache.bcel.verifier.structurals.UninitializedObjectType;
import org.apache.bcel.util.Repository;
import org.apache.bcel.generic.*;
import org.apache.bcel.classfile.*;

import edu.purdue.cs.bloat.editor.ClassHierarchy;
import edu.purdue.cs.bloat.editor.MemberRef;
import edu.purdue.cs.bloat.editor.MethodRef;
import edu.purdue.cs.bloat.editor.NameAndType;

import java.util.*;

/**
 * Maintains all BLOAT data structures as if they were meant to reside in a
 * persistent store. As a result, it keeps every piece of BLOAT data around
 * because it might be needed in the future. No fancing cache maintainence is
 * performed. Because we are going for maximum information we take the closure
 * of classes when working with the class hierarchy.
 */
public class PersistentBloatContext extends BloatContext {

    protected final ClassHierarchy hierarchy;

    protected Map classEditors; // Maps ClassInfos to ClassEditors

    protected Map classInfos; // Maps ClassInfos to ClassEditors

    public static boolean DB_COMMIT = true;

    protected static void comm(String s) {
        if (DB_COMMIT || DEBUG) {
            System.out.println(s);
        }
    }

    /**
     * Constructor. Each <tt>BloatContext</tt> stems from a
     * <tt>ClassInfoLoader</tt>. Using the loader it can create an
     * <tt>Editor</tt> and such. Initially, no classes are loaded.
     */
    public PersistentBloatContext(Repository loader) {
        this(loader, true);
    }

    /**
     * Constructor. It is the responsibility of the subclasses to add classes to
     * the hierarchy by calling <tt>addClasses</tt>.
     * 
     * @param loader
     *            Used to load classes
     * @param closure
     *            Do we look for the maximum number of classes?
     */
    public PersistentBloatContext(Repository loader, boolean closure) {
        super(loader);
        db("Creating a new BloatContext");

        // Create a bunch of the mappings we maintain. Make sure to do
        // this before anything else!

        classEditors = new LinkedHashMap();
        classInfos = new LinkedHashMap();

        // Have to create an empty class hierarchy then add the classes.
        // There is a strange circular dependence between the hierarchy
        // and the context.
        this.hierarchy = new ClassHierarchy(this, new ArrayList(), closure);
    }

    /**
     * Adds a bunch of (names of) classes to the hierarchy.
     */
    protected void addClasses(Collection classes) {
        Iterator iter = classes.iterator();
        while (iter.hasNext()) {
            String className = (String) iter.next();
            this.hierarchy.addClassNamed(className);
        }
    }

    public JavaClass loadClass(String className) throws ClassNotFoundException {
        // Lots of interesting stuff to do here. For the moment, just
        // load the class from the ClassInfoLoader and add it to the
        // hierarchy.

        className = className.replace('.', '/');

        // Check the cache of ClassInfos
        JavaClass info = (JavaClass) classInfos.get(className);

        if (info == null) {
            db("BloatContext: Loading class " + className);
            info = loader.loadClass(className);
            hierarchy.addClassNamed(className);
            db("loadClass: " + className + " -> " + info);
            classInfos.put(className, info);
        }

        return (info);
    }

    public JavaClass newClassInfo(int class_name_index,
            int superclass_name_index, String file_name, int major, int minor,
            int access_flags, ConstantPool constant_pool, int[] interfaces,
            Field[] fields, Method[] methods, Attribute[] attributes) {
        JavaClass jc = new JavaClass(class_name_index, superclass_name_index,
                file_name, major, minor, access_flags, constant_pool,
                interfaces, fields, methods, attributes);
        String className = jc.getClassName();

        db("newClassEditor(...........): " + className + " -> " + jc);

        classInfos.put(className, jc);
        return jc;
    }

    public ClassGen newClass(int accessflags, String className,
            String file_name, Type superType, Type[] interfaces) {
        String[] interNames = new String[interfaces.length];
        for (int i = 0; i < interfaces.length; i++) {
            interNames[i] = ((ObjectType) interfaces[i]).getClassName();
        }

        ClassGen cg = new ClassGen(className, ((ObjectType) superType)
                .getClassName(), file_name, accessflags, interNames);
        JavaClass jc = cg.getJavaClass();
        String classname = cg.getClassName();
        db("newClassInfo(.....): " + classname + " -> " + cg);
        classEditors.put(classname, cg);
        classInfos.put(classname, jc);
        return cg;
    }

    /*
     * public ClassInfo newClassInfo(int modifiers, int classIndex, int
     * superClassIndex, int[] interfaceIndexes, List constants) {
     * 
     * return this.loader.storeClass(modifiers, classIndex, superClassIndex,
     * interfaceIndexes, constants); }
     */

    public ClassHierarchy getHierarchy() {
        return (this.hierarchy);
    }

    public ClassGen newClass(int class_name_index, int superclass_name_index,
            String file_name, int major, int minor, int access_flags,
            ConstantPool constant_pool, int[] interfaces, Field[] fields,
            Method[] methods, Attribute[] attributes) {
        JavaClass jc = new JavaClass(class_name_index, superclass_name_index,
                file_name, major, minor, access_flags, constant_pool,
                interfaces, fields, methods, attributes);

        ClassGen ce = new ClassGen(jc);

        String className = ce.getClassName();

        db("editClass(ClassInfo): " + className + " -> " + ce);

        classEditors.put(className, ce);
        classInfos.put(className, jc);
        return ce;
    }

    public ClassGen editClass(String className) throws ClassNotFoundException,
            ClassFormatException {
        // Only make the name -> classInfo mapping if we edit the class,
        // this way the mapping will be deleted when the ClassEditor is
        // released.

        // className = className.intern();

        ClassGen info = (ClassGen) classEditors.get(className);

        if (info == null) {
            JavaClass jc = loadClass(className);
            db("editClass(String): " + className + " -> " + info);
            return editClass(jc);
        }

        return (info);
    }

    public ClassGen editClass(Type classType) throws ClassNotFoundException,
            ClassFormatException {
        if (classType instanceof ReferenceType) {
            if (classType instanceof ObjectType)
                return editClass((ObjectType) classType);
            else if (classType instanceof ArrayType)
                return editClass(((ArrayType) classType).getBasicType());
            else if (classType instanceof UninitializedObjectType)
                return editClass(((UninitializedObjectType) classType)
                        .getInitialized());
        } else
            throw new ClassNotFoundException(classType
                    + "is not an ObjectType and cannot be loaded");
        return null;
    }

    public ClassGen editClass(ObjectType classType)
            throws ClassNotFoundException, ClassFormatException {
        return (editClass(classType.getClassName()));
    }

    public ClassGen editClass(JavaClass info) {
        // Check the cache
        ClassGen ce = (ClassGen) classEditors.get(info.getClassName());

        if (ce == null) {
            ce = new ClassGen(info);
            classInfos.put(info.getClassName(), info);
            classEditors.put(info.getClassName(), ce);

            // if(!classInfos.containsValue(info)) {
            // String className = ce.getClassName().intern();
            // db("editClass(ClassInfo): " + className + " -> " + info);
            // classInfos.put(className, info);
            // }
        }

        return (ce);
    }

    public MethodGen editMethod(MethodRef method) throws NoSuchMethodException {

        // Groan, we have to do this the HARD way.
        db("Creating a new MethodEditor for " + method);
        String name = method.name();
        String signature = method.getSignature();

        try {
            ClassGen ce = editClass(method.declaringClass());
            Method[] methods = ce.getMethods();

            for (int i = 0; i < methods.length; i++) {
                MethodGen me = new MethodGen(methods[i], method
                        .declaringClass().getSignature(), ce.getConstantPool());// TODO

                if (me.getName().equals(name)
                        && me.getSignature().equals(signature)) {
                    // The call to editMethod should have already handled
                    // the methodEditors mapping, but we still need to do
                    // methodInfos.
                    return (me);
                }

                // release(methods[i], ce.getClassName());
            }

        } catch (ClassNotFoundException ex1) {
        } catch (ClassFormatException ex2) {
        }

        throw new NoSuchMethodException(method.toString());
    }

    public MethodGen editMethod(Method info, String declaringClass) {
        // Groan, we have to do this the HARD way.
        db("Creating a new MethodEditor for " + info.getName());
        String name = info.getName();
        String signature = info.getSignature();

        try {
            ClassGen ce = editClass(declaringClass);
            Method[] methods = ce.getMethods();

            for (int i = 0; i < methods.length; i++) {

                if (methods[i].getName().equals(name)
                        && methods[i].getSignature().equals(signature)) {
                    MethodGen me = new MethodGen(methods[i], declaringClass, ce
                            .getConstantPool());
                    // The call to editMethod should have already handled
                    // the methodEditors mapping, but we still need to do
                    // methodInfos.
                    return (me);
                }

                // release(methods[i], ce.getClassName());
            }

        } catch (ClassNotFoundException ex1) {
        } catch (ClassFormatException ex2) {
        }
        return null;
        // throw new NoSuchMethodException(info.toString());
    }

    public FieldGen editField(MemberRef field) throws NoSuchFieldException {

        // Just like we had to do with methods

        NameAndType nat = field.nameAndType();
        String name = nat.name();
        Type type = nat.type();

        try {
            ClassGen ce = editClass(field.declaringClass());
            Field[] fields = ce.getFields();

            for (int i = 0; i < fields.length; i++) {
                FieldGen fe = editField(fields[i], field.declaringClass()
                        .getSignature());// TODO

                if (fe.getName().equals(name) && fe.getType().equals(type)) {
                    return (fe);
                }

            }
        } catch (ClassNotFoundException ex1) {
        } catch (ClassFormatException ex2) {
        }
        return null;
        // throw new NoSuchFieldException(field.toString());
    }

    public FieldGen editField(Field info, String className) {
        // Check the cache
        String name = info.getName();
        Type type = info.getType();

        try {
            ClassGen ce = editClass(className);
            Field[] fields = ce.getFields();

            for (int i = 0; i < fields.length; i++) {
                FieldGen fe = editField(fields[i], className);

                if (fe.getName().equals(name) && fe.getType().equals(type)) {
                    return (fe);
                }

            }
        } catch (ClassNotFoundException ex1) {
        } catch (ClassFormatException ex2) {
        }
        return null;
        // throw new NoSuchFieldException(info.toString());

    }

    public void release(JavaClass info) {
        // Since we keep around all data, do nothing
    }

    public void release(ClassGen ce) {
        // Since we keep around all data, do nothing
    }

    public void release(Method info, String className) {
        // Since we keep around all data, do nothing
    }

    public void release(Field info, String className) {
        // Since we keep around all data, do nothing
    }

    /**
     * Classes that are ignored are not committed.
     * 
     * @see ignoreClass
     */
    public void commit(JavaClass info) {
        ClassGen cg = (ClassGen) classEditors.get(info.getClassName());
        JavaClass jc = (JavaClass) cg.getJavaClass();
        classInfos.put(info.getClassName(), jc);
    }

    public void commit(ClassGen info) {
        classEditors.put(info.getClassName(), info);
    }

    public void commit(MethodGen methodGen) {
        this.commit(methodGen.getMethod(), methodGen.getClassName());
    }

    public void commit(Method m, String classname) {
        try {
            ClassGen cg = editClass(classname);
            Method oldMethod = cg.containsMethod(m.getName(), m.getSignature());
            cg.replaceMethod(oldMethod, m);
        } catch (ClassNotFoundException cnfe) {
            // IF THERE IS NO CLASS THERE IS NOTHING TO COMMIT, SHOULDN'T
            // HAPPEN
        }
    }

    /*
     * private Method commit(Method m){ try{ ClassGen cg = editClass(classname);
     * Method oldMethod = cg.containsMethod(m.getName(),m.getSignature());
     * cg.replaceMethod(oldMethod, m); }catch(ClassNotFoundException cnfe){ //IF
     * THERE IS NO CLASS THERE IS NOTHING TO COMMIT, SHOULDN'T HAPPEN } }
     */
    public void commit(Field f, String classname) {
        try {
            ClassGen cg = editClass(classname); // Ask to commit the ClassGen
            Field oldField = cg.containsField(f.getName());
            cg.replaceField(oldField, f);
        } catch (ClassNotFoundException cnfe) {
            // IF THERE IS NO CLASS THERE IS NOTHING TO COMMIT, SHOULDN'T
            // HAPPEN
        }
    }

    /*
     * private Field commit(Field f){ FieldGen fg =
     * (FieldGen)fieldEditors.get(f); if(fg != null){ fg.update(); Field
     * newField = fg.getField(); //return final version of fg
     * fieldEditors.remove(f); fieldEditors.put(newField, fg); return newField;
     * }else{ return f; //return the old field } }
     */

    // THE COMMIT OF A METHOD/FIELD SHOULD RESULT IN THE COMMIT OF THE
    // PARENT
    // CLASS SO commit(JavaClass) SHOULD SUFFICE.
    // Reason I got rid of commit(method) is bcel's Method has no mapping to
    // the class that the method belongs to. Each editField or editMethod
    // ensures that the class is loaded. (release does nothing)
    public void commit() {
    }

    public void commitDirty() {
    }

    /*
     * // Commit all dirty fields Object[] array =
     * this.fieldEditors.values().toArray(); for(int i = 0; i < array.length;
     * i++) { FieldEditor fe = (FieldEditor) array[i]; if(fe.isDirty() &&
     * !ignoreField(fe.memberRef())) { comm(" Committing field: " +
     * fe.declaringClass().name() + "." + fe.name()); commit(fe.fieldInfo()); } } //
     * Commit all dirty methods array = this.methodEditors.values().toArray();
     * for(int i = 0; i < array.length; i++) { MethodEditor me = (MethodEditor)
     * array[i]; if(me.isDirty() && !ignoreMethod(me.memberRef())) { comm("
     * Committing method: " + me.declaringClass().name() + "." + me.name() +
     * me.type()); commit(me.methodInfo()); } } // Commit all dirty classes
     * array = this.classEditors.values().toArray(); for(int i = 0; i <
     * array.length; i++) { ClassEditor ce = (ClassEditor) array[i];
     * if(ce.isDirty() && !ignoreClass(ce.type())) { comm(" Committing class: " +
     * ce.name()); commit(ce.classInfo()); } } }
     */

}
