/*
 * Class: Inline 
 * Version: 
 * Date:
 * 
 * All files in the distribution of BLOAT (Bytecode Level Optimization and
 * Analysis tool for Java(tm)) are Copyright 1997-2001 by the Purdue
 * Research Foundation of Purdue University.  All rights reserved.
 *
 * <p>
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this entire copyright notice is duplicated in all
 * such copies, and that any documentation, announcements, and other
 * materials related to such distribution and use acknowledge that the
 * software was developed at Purdue University, West Lafayette, IN by
 * @author Antony Hosking
 * @author David Whitlock
 * @author Nathaniel Nystrom
 * No charge may be made for copies, derivations, or distributions of
 * this material without the express written consent of the copyright
 * holder.  Neither the name of the University nor the name of the
 * author may be used to endorse or promote products derived from this
 * material without specific prior written permission.  THIS SOFTWARE
 * IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR ANY PARTICULAR PURPOSE.
 *
 * <p>
 * 
 * Port of the BLOAT to use the Jakata Apache project BCEL 5.2 was 
 * contributed by The Australian National University by
 * @author Arrin Daley
 * @author John N Zigman
 * 
 * <p>
 * Java is a trademark of Sun Microsystems, Inc.
 */

package edu.purdue.cs.bloat.inline;

import java.util.*;

import org.apache.bcel.generic.*;

import edu.purdue.cs.bloat.cfg.FlowGraph;
import edu.purdue.cs.bloat.editor.MethodRef;
import edu.purdue.cs.bloat.editor.NameAndType;
import edu.purdue.cs.bloat.util.Assert;

/**
 * Inlines the code of non-virtual method call sites. These sites include calls
 * to static methods and certain uses of the <tt>invokespecial</tt> method.
 * There are certain metrics that can be set to effect where and how inlining is
 * performed.
 */
public class Inline {
    public static boolean DEBUG = false;

    private int maxCodeSize; // Max number of instructions in method

    private int maxInlineSize; // Don't inline methods larger than this

    private int maxCallDepth; // Max of height of call stack

    private boolean inlineExceptions; // Inline methods that throw exceptions

    private InlineContext context;

    /**
     * Size of the largest method that can be inlined
     */
    public static int CALLEE_SIZE = 100000;

    private static void db(String s) {
        if (DEBUG) {
            System.out.println(s);
        }
    }

    /**
     * Constructor. By default the first-level calls are only inlined one level
     * deep, there is no max size on methods to inline, and methods that may
     * throw exceptions are inlined.
     * 
     * @param maxCodeSize
     *            The maximum number of instructions a method can grow to.
     */
    public Inline(InlineContext context, int maxCodeSize) {
        this.context = context;
        this.maxCodeSize = maxCodeSize;
        this.maxInlineSize = -1;
        this.maxCallDepth = 1;
        this.inlineExceptions = true;
    }

    /**
     * Sets the maximum of size of a method that will be inlined. No method
     * larger than this will be inlined.
     */
    public void setMaxInlineSize(int maxInlineSize) {
        this.maxInlineSize = maxInlineSize;
    }

    /**
     * Sets the maximum number of nested calls we inline.
     */
    public void setMaxCallDepth(int maxCallDepth) {
        this.maxCallDepth = maxCallDepth;
    }

    /**
     * Sets whether or not methods that may throw exceptions (that is, have a
     * non-empty "throws" declaration) are inlined.
     */
    public void setInlineExceptions(boolean inlineExceptions) {
        this.inlineExceptions = inlineExceptions;
    }

    /**
     * Scans a method and inlines non-virtual method calls according to this
     * <tt>Inline</tt>'s metrics.
     */
    public void inline(MethodGen method) {
        ConstantPoolGen cpg = method.getConstantPool();
        // Go through the method and look for calls to inline
        StackHeightCounter stackHeight = new StackHeightCounter(method);
        InstructionList code = method.getInstructionList();
        FlowGraph.buildBLOATInstructionList(method);
        boolean firstCall = true;
        Iterator iter = code.iterator();
        while (iter.hasNext()) {
            InstructionHandle ih = (InstructionHandle) iter.next();
            if (FlowGraph.label(ih)) {
                // stackHeight.handle(ih);
                // All instructions get handled later
                db("  " + ih + "." + stackHeight.height() + ") " + ih
                        + (FlowGraph.label(ih) ? " (starts block)" : ""));
            }

            Instruction inst = ih.getInstruction();
            if (inst instanceof INVOKESTATIC || inst instanceof INVOKESPECIAL) {
                InvokeInstruction iI = (InvokeInstruction) inst;
                MethodRef callee = new MethodRef((ObjectType) iI
                        .getReferenceType(cpg), new NameAndType(iI
                        .getMethodName(cpg), iI.getReturnType(cpg)), iI
                        .getReturnType(cpg), iI.getArgumentTypes(cpg));
                Stack callStack = new Stack();

                callStack.add(new MethodRef(new ObjectType(method
                        .getClassName()), new NameAndType(method.getName(),
                        method.getReturnType()), method.getReturnType(), method
                        .getArgumentTypes()));

                db("  Call: " + inst);

                stackHeight.handle(ih);
                int expectedHeight = stackHeight.height();
                stackHeight.unhandle(ih);

                InstructionHandle jh = ih;
                ih = inline(method, callee, ih, callStack, stackHeight,
                        firstCall);

                if (jh == ih) {
                    // Call was not inlined, add it to the stack
                    stackHeight.handle(ih);
                    db("  " + ih + "." + stackHeight.height() + ") " + inst);
                }

                int newHeight = stackHeight.height();
                // If an exception is thrown as the last thing in a method
                // newHeight will equal 0. Let's let this one slide.
                Assert.isTrue(newHeight == 0 || newHeight == expectedHeight,
                        "Inlining did not get the stack heights right: "
                                + "Expected " + expectedHeight + ", got "
                                + newHeight);

            } else {
                stackHeight.handle(ih);
                db("  " + ih + "." + stackHeight.height() + ") " + inst);
            }

            if (inst instanceof InvokeInstruction) {
                firstCall = false;
            }

        }

        // method.setCode(code); WE JUST MODIFIED IT SO CHANGES ARE COMMITED

        if (DEBUG) {
            stackHeight = new StackHeightCounter(method);
            db("\nNew Code for " + method.getClassName() + "."
                    + method.getName() + method.getReturnType());
            code = method.getInstructionList();
            Iterator iterator = method.getInstructionList().iterator();
            InstructionHandle j = null;
            while (iterator.hasNext()) {
                j = (InstructionHandle) iterator.next();
                if (FlowGraph.label(j)) {

                    // stackHeight.handle(j);

                    CodeExceptionGen[] tryCatches = method
                            .getExceptionHandlers();
                    for (int i = 0; i < tryCatches.length; i++) {
                        CodeExceptionGen tryCatch = tryCatches[i];
                        if (tryCatch.getStartPC().equals(j)) {
                            System.out.println(" Begin protected region");
                        }

                        if (tryCatch.getEndPC().equals(j)) {
                            System.out.println(" End protected region");
                        }

                        // A Label can both end a protected region and begin
                        // catch
                        // block

                        if (tryCatch.getHandlerPC().equals(j)) {
                            System.out.println(" Catch "
                                    + tryCatch.getCatchType());
                        }

                    }

                }
                stackHeight.handle(j);
                System.out.println("  " + j + "." + stackHeight.height() + ") "
                        + j.getInstruction()
                        + (FlowGraph.label(j) ? " (starts block)" : ""));

            }

            // Print try-catch information
            CodeExceptionGen[] tryCatches = method.getExceptionHandlers();
            System.out.println("Exception information:");
            for (int i = 0; i < tryCatches.length; i++) {
                System.out.println("  " + tryCatches[i]);
            }
            System.out.println("");
        }

    }

    /**
     * Helper method that does most of the work. By calling this method
     * recursively, we can inline more than one call deep.
     * 
     * @param caller
     *            The original caller that got all of this started. Into this
     *            method we insert the code.
     * @param callee
     *            The method to be inlined
     * @param index
     *            Where in caller we insert inlined code
     * @param callStack
     *            A stack of <tt>MemberRef</tt>s that represent the inlined
     *            methods that call other methods. It is used to detect
     *            recursion.
     * 
     * @return The index into the caller's code array of the instruction
     *         following the last inlinined instruction. Start looking here
     *         after inline returns.
     */
    // isRecursiveCount added so I can be sure of getting a unique local
    // variable name
    private InstructionHandle inline(MethodGen caller, MethodRef callee,
            InstructionHandle index, Stack callStack,
            StackHeightCounter stackHeight, boolean firstCall) {

        Instruction newInst = null;
        int LOCALVARINDEX = caller.getMaxLocals() - 1;

        // Do we ignore the method being inlined?
        if (context.ignoreMethod(callee)) {
            db("  Can't inline " + callee + ": it's ignored");
            index = index.getNext();
            return (index);
        }

        // Can we inline this method
        if (callStack.size() > maxCallDepth) {
            db("  Can't inline " + callee + ": max call depth (" + maxCallDepth
                    + ") reached");
            index = index.getNext();
            return (index);

        } else if (callStack.contains(callee)) {
            db("  Can't inline recursive call to " + callee);
            index = index.getNext();
            return (index);
        }

        // Make sure we're not inlining the static-ized version of a
        // method in the call stack.
        String name = callee.name();
        // Is this stuff that BCEL would do? or has a BLOAT optimization done it
        int b = name.indexOf("$$BLOAT");
        if (b != -1) {
            name = name.substring(0, b);

            // Get rid of first parameter
            Type[] oldParams = callee.paramTypes();
            Type[] newParams = new Type[oldParams.length - 1];
            for (int p = 1; p < oldParams.length; p++) {
                newParams[p - 1] = oldParams[p];
            }

            MethodRef unBloated = new MethodRef(callee.declaringClass(),
                    new NameAndType(name, callee.returnType()), callee
                            .returnType(), newParams);

            if (callStack.contains(unBloated)) {
                db("  Can't inline recursive call to " + callee);
                index = index.getNext();
                return (index);
            }
        }

        InstructionList code = caller.getInstructionList();
        if (code.size() > maxCodeSize) {
            db("  Can't inline " + callee + ": max code size (" + maxCodeSize
                    + ") reached");
            index = index.getNext();
            return (index);
        }

        MethodGen calleeMethod = null;
        try {
            calleeMethod = context.editMethod(callee);

        } catch (NoSuchMethodException ex) {
            System.err.println("Couldn't find method " + callee);
            ex.printStackTrace(System.err);
            System.exit(1);
        }

        if (calleeMethod.isNative()) {
            db("  Can't inline " + callee + ": it's a native method");
            index = index.getNext();
            return (index);
        }

        if (calleeMethod.isSynchronized()) {
            db("  Can't inline " + callee + ": it's synchronized");
            index = index.getNext();
            return (index);
        }

        if (!inlineExceptions && calleeMethod.getExceptions().length > 0) {
            db("  Can't inline " + callee + ": it may throw an exception");
            index = index.getNext();
            return (index);
        }

        if (calleeMethod.getInstructionList().size() > maxInlineSize) {
            db("  Can't inline " + callee + ": it's too big");
            index = index.getNext();
            return (index);
        }

        // Methods that catch exceptions are problematic. When an
        // exception is thrown, it clears the stack. Ordinarily this
        // isn't a problem. However, now the stack of the caller is
        // cleared in addition to the stack of the callee. This is bad.
        // The callee might catch the exception and deal with it.
        // However, the stack has still been cleared. This really messes
        // things up for the code that appears after the inlined method.
        // So, if a method catches an exception, we can only inline it if
        // the stack contains nothing but the parameters to the call.
        if (calleeMethod.getExceptionHandlers().length > 0) {
            int size = 0;
            Type[] args = callee.paramTypes();
            for (int i = 0; i < args.length; i++) {
                size = size + args[i].getSize();
            }
            if (stackHeight.height() > size) {
                db("  Can't inline " + callee
                        + ": It catches an exception and there's stuff on the "
                        + "stack");
                index = index.getNext();
                return (index);
            }
        }

        // If the callee method catches any of the same exceptions as the
        // protected region that we are currently in, then we can't inline
        // the method.
        CodeExceptionGen[] tryCatches0 = calleeMethod.getExceptionHandlers();
        for (int i = 0; i < tryCatches0.length; i++) {
            CodeExceptionGen tc1 = tryCatches0[i];
            Iterator iter = stackHeight.tryCatches().iterator();
            while (iter.hasNext()) {
                CodeExceptionGen tc2 = (CodeExceptionGen) iter.next();
                Type t1 = tc1.getCatchType();
                Type t2 = tc2.getCatchType();
                if (t1 != null && t2 != null && t1.equals(t2)) {
                    db("  Can't inline " + callee
                            + ": It catches the same type "
                            + tc1.getCatchType().getClassName()
                            + " as the current protected region");
                    index = index.getNext();
                    return (index);
                }
            }
        }

        // If the caller is a constructor and this is the first
        // invokespecial we've seen in this callee method, we can inline
        // calls to the constructors of superclasses and other
        // constructors in this method. So, if this IS the first call in
        // a method which IS a constructor, we can inline it.
        if (calleeMethod.getName().equals("<init>")
                && (!firstCall || !caller.getName().equals("<init>"))) {

            db("  Can't inline " + callee + ": It calls a normal constructor");
            index = index.getNext();
            return (index);
        }

        // Local variables are problematic. We cannot simply map the
        // callee's variables to new variables in the caller because we
        // cannot precisely determine the width of the variable the first
        // time we see it. (For instance, Nate's generated code might use
        // a local variable as a non-wide initially and then use it as a
        // wide later.)

        // Okay, we going to inline. Remove the calling instruction.
        Instruction call = index.getInstruction();
        try {
            code.delete(index);
        } catch (TargetLostException tle) {
            // TargetLost? Oh well, go on with inlining
        }
        index = index.getPrev();
        db("  Removing call: " + call);
        Assert.isTrue(call instanceof INVOKESTATIC
                || call instanceof INVOKESPECIAL,
                "Removing the wrong call instruction:" + call);
        callStack.push(callee);

        db("  Inlining call (" + callStack.size() + ") to "
                + callee.declaringClass() + "." + callee.name() + callee.type());
        context.getInlineStats().noteInlined();

        // ALL RIGHT, CREATE LOCALVARAIBLES FOR EACH OF THE LOCALS IN THE CALLEE
        // The index should be LOCALVARINDEX + the index set in the calle code
        // These will be added where they are params or garden variety locals.
        // indexes shouldn't be a problem.
        LocalVariableGen[] locals = calleeMethod.getLocalVariables();
        for (int i = 0; i < locals.length; i++) {
            caller.addLocalVariable("inline" + calleeMethod.getName() + i,
                    locals[i].getType(), index, index.getNext());
        }

        // First we have to pop the arguments off the stack and store them
        // into the local variables. Remember that wide types occupy two
        // local variables.

        // Mapper mapper = new Mapper(caller);
        // Type[] paramTypes = callee.paramTypes();

        // IS THERE ANY REASON TO GO BACKWARDS?
        for (int i = locals.length - 1; i >= 0; i--) {
            // Map the local variables containing the arguments to new
            // local variables.
            Type paramType = locals[i].getType();

            if (paramType == null) {
                continue;
            }

            db("  Param " + i + ": of type " + paramType);

            // LocalVariable newVar = mapper.map(param, paramType);
            // I've already created the new LocalVariable

            if (paramType instanceof ReferenceType) {
                newInst = new ASTORE(locals[i].getIndex());
            } else {
                if (paramType.equals(Type.BOOLEAN)
                        || paramType.equals(Type.BYTE)
                        || paramType.equals(Type.CHAR)
                        || paramType.equals(Type.SHORT)
                        || paramType.equals(Type.INT)) {
                    newInst = new ISTORE(locals[i].getIndex());
                } else if (paramType.equals(Type.DOUBLE)) {
                    newInst = new DSTORE(locals[i].getIndex());
                } else if (paramType.equals(Type.LONG)) {
                    newInst = new LSTORE(locals[i].getIndex());
                } else if (paramType.equals(Type.FLOAT)) {
                    newInst = new FSTORE(locals[i].getIndex());
                } else {
                    throw new IllegalArgumentException("What's a " + paramType
                            + "doing as a method " + "parameter");
                }

            }

            index = caller.getInstructionList().append(index, newInst);
            stackHeight.handle(index);
            db("  " + index + "." + stackHeight.height() + "> " + newInst);
        }
        // SO ALL THE LOCALS ARE SETUP

        // Before we mess with the code, we have to patch up the try-catch
        // information from the inlined method to the caller method.
        CodeExceptionGen[] tryCatches = calleeMethod.getExceptionHandlers();
        for (int i = 0; i < tryCatches.length; i++) {
            CodeExceptionGen tryCatch = tryCatches[i];

            // INSTRUCTIONHANDLES STAY INTACT WHEN A INSTRUCTIONLIST IS
            // ADDED
            // SO WE DON"T NEED TO TRANSLATE LABELS AS BLOAT DID
            InstructionHandle start = tryCatch.getStartPC();
            InstructionHandle end = tryCatch.getEndPC();
            InstructionHandle handler = tryCatch.getHandlerPC();

            caller.addExceptionHandler(start, end, handler, tryCatch
                    .getCatchType());
        }

        // Go through the code in the callee method and inline it. Handle
        // any calls by making a recursive call to this method. Copy each
        // instruction to the method in which it is being inlined. Along
        // the way convert references to local variables to their mapped
        // values. Also remove return instructions. Replace them with
        // loads as necessary.

        FlowGraph.buildBLOATInstructionList(calleeMethod);
        InstructionList inlineCode = calleeMethod.getInstructionList().copy();
        // I want to play with the instructionlist and then add it to the caller
        // code in one go which saves me messing with mapping instructionHandles

        // We don't want to introduce a new end label because it confuses
        // BLOAT during CFG construction. We designate the end label as
        // starting a new block in hopes that it will solve problems with
        // CFG construction.
        InstructionHandle last = inlineCode.append(new NOP());
        InstructionHandle endLabel;
        if (FlowGraph.label(last)) {
            endLabel = last;
        } else {
            FlowGraph.setLabel(last);
            endLabel = last;
        }
        // endLabel.setStartsBlock(true);

        firstCall = true;
        Iterator iter = inlineCode.iterator();
        while (iter.hasNext()) {
            InstructionHandle instHandle = (InstructionHandle) iter.next();
            Instruction inst = instHandle.getInstruction();

            if (FlowGraph.label(instHandle)) {
                // code.add(++index, newLabel);
                // stackHeight.handle(newLabel);
                db("  "
                        + index
                        + "."
                        + stackHeight.height()
                        + "> "
                        + inst
                        + (FlowGraph.label(instHandle) ? " (starts block)" : ""));
            }

            // Assert.isTrue(o instanceof Instruction, "What is a " + o +
            // " doing in the instruction stream?");

            // int opcode = inst.opcodeClass();

            if (inst instanceof LocalVariableInstruction) {
                // Map local variable in the callee method to local
                // variables in the caller method.
                // LocalVariable local = mapper.map((LocalVariable) operand,
                // (inst.category() == 2 ? true : false));
                // operand = local;
                LocalVariableInstruction lvi = (LocalVariableInstruction) inst;
                lvi.setIndex(lvi.getIndex() + LOCALVARINDEX);

                // } else if(inst instanceof Label) {
                // Map labels in the callee method to labels in the caller
                // method.
                // Label label = mapper.map((Label) operand);
                // operand = label;
                // HOPEFULLY THE InstructionHandles STAY INTACT

            } else if (inst instanceof Select) {
                // We have to patch up the Labels involved with the Switch
                // Switch oldSwitch = (Switch) operand;

                // Label newDefault = mapper.map(oldSwitch.defaultTarget());

                // Label[] oldTargets = oldSwitch.targets();
                // Label[] newTargets = new Label[oldTargets.length];
                // for(int i = 0; i < newTargets.length; i++) {
                // Label newTarget = mapper.map(oldTargets[i]);
                // newTargets[i] = newTarget;
                // }

                // operand = new Switch(newDefault, newTargets,
                // oldSwitch.values());
                // HOPEFULLY THE InstructionHandles STAY INTACT
            }

            if (inst instanceof ReturnInstruction) {
                // Insert a jump to the end of the inlined method. Any
                // return value will be on top of the stack. This is where
                // we want it.
                newInst = new GOTO(endLabel);
                instHandle.setInstruction(newInst);
                // code.append(index, newInst);
                index = index.getNext();
                stackHeight.handle(instHandle);
                db("  " + index + "." + stackHeight.height() + "> " + newInst);

            } else if (inst instanceof INVOKESTATIC
                    || inst instanceof INVOKESPECIAL) {
                // Make a recursive call. Note that this must be done after
                // we add the call instruction above. But we only want to
                // visit the instruction with the stackHeight if the call was
                // not inlined.
                // newInst = new Instruction(opcode, operand);
                // code.add(++index, newInst);
                InvokeInstruction iI = (InvokeInstruction) inst;

                stackHeight.handle(instHandle);
                int expectedHeight = stackHeight.height();
                stackHeight.unhandle(instHandle);
                ConstantPoolGen cpg = calleeMethod.getConstantPool();
                MethodRef nestedCall = new MethodRef((ObjectType) iI
                        .getReferenceType(cpg), new NameAndType(iI
                        .getMethodName(cpg), iI.getReturnType(cpg)), iI
                        .getReturnType(cpg), iI.getArgumentTypes(cpg));
                InstructionHandle oldIndex = instHandle;
                instHandle = inline(caller, nestedCall, instHandle, callStack,
                        stackHeight, firstCall);

                if (instHandle == oldIndex) {
                    stackHeight.handle(instHandle);
                    db("  " + instHandle + "." + stackHeight.height() + "> "
                            + inst);
                }

                int newHeight = stackHeight.height();
                Assert.isTrue(newHeight == 0 || newHeight == expectedHeight,
                        "Inlining did not get the stack heights right: "
                                + "Expected " + expectedHeight + ", got "
                                + newHeight);

            } else {
                // Add the instruction
                // newInst = new Instruction(opcode, operand);
                // code.add(++index, newInst);
                stackHeight.handle(instHandle);
                db("  " + index + "." + stackHeight.height() + "> " + newInst);
            }

            // We want to do this after we've made any recursive calls to
            // inline.
            if (inst instanceof InvokeInstruction) {
                firstCall = false;
            }

        }

        code.append(index, inlineCode);
        index = endLabel.getNext();
        // add the whole of the inline method code now that it has been modified

        // if(addEndLabel) {
        // Done inlining. Add end label.
        // code.add(++index, endLabel);
        // stackHeight.handle(endLabel);
        // db(" " + index + "." + stackHeight.height() + "> " + endLabel +
        // (endLabel.startsBlock() ? " (starts block)" : ""));
        // }

        // caller.setDirty(true);
        callStack.pop();
        return (index);

    }
}

/**
 * Utility class for mapping local variables and labels. Note that when mapping
 * local variables we have to be careful. We can't assume that a variable will
 * retain its "wideness" throughout the method. I learned this one the hard way.
 * So, we have to keep a constant difference between the mapped variables.
 */
/*
 * class Mapper { private Map varsMap; // Maps local variables //private Map
 * labelsMap; // Maps labels private MethodGen method; // Method into which
 * things are mapped //private int offset; // Start numbering new locals here
 * //adding locals to method will do this for us.
 * 
 * private static void db(String s) { if(Inline.DEBUG) System.out.println(s); }
 */

/**
 * Constructor.
 */
/*
 * public Mapper(MethodGen method) { this.method = method; varsMap = new
 * HashMap(); labelsMap = new LinkedHashMap(); offset = method.maxLocals() + 1; }
 * 
 * public Label map(Label label) { Label newLabel = (Label)
 * labelsMap.get(label); if(newLabel == null) { newLabel =
 * this.method.newLabel(); newLabel.setStartsBlock(label.startsBlock());
 * labelsMap.put(label, newLabel); db(" " + label + " -> " + newLabel +
 * (newLabel.startsBlock() ? " (starts block)" : "")); } return(newLabel); }
 * 
 * public LocalVariableGen map(LocalVariableGen var, Type type) {
 * LocalVariableGen newVar = (LocalVariableGen) varsMap.get(var); if(newVar ==
 * null) { newVar = this.method.localAt(var.index() + offset); // newVar =
 * this.method.newLocal(type); varsMap.put(var, newVar); db(" " + var + " (" +
 * var.index() + ") -> " + newVar + "(" + var.index() + "+" + offset + ")" +
 * (type.isWide() ? " (" + type + ")" : "")); } return(newVar); }
 * 
 * public LocalVariable map(LocalVariable var, boolean isWide) { LocalVariable
 * newVar = (LocalVariable) varsMap.get(var); if(newVar == null) { newVar =
 * this.method.localAt(var.index() + offset); // newVar =
 * this.method.newLocal(isWide); varsMap.put(var, newVar); db(" " + var + " (" +
 * var.index() + ") -> " + newVar + "(" + var.index() + "+" + offset + ")" +
 * (isWide ? " (wide)" : "")); } return(newVar); } }
 */

