/*
 * Class: StackHeightCounter 
 * Version: 
 * Date:
 * 
 * All files in the distribution of BLOAT (Bytecode Level Optimization and
 * Analysis tool for Java(tm)) are Copyright 1997-2001 by the Purdue
 * Research Foundation of Purdue University.  All rights reserved.
 *
 * <p>
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this entire copyright notice is duplicated in all
 * such copies, and that any documentation, announcements, and other
 * materials related to such distribution and use acknowledge that the
 * software was developed at Purdue University, West Lafayette, IN by
 * @author Antony Hosking
 * @author David Whitlock
 * @author Nathaniel Nystrom
 * No charge may be made for copies, derivations, or distributions of
 * this material without the express written consent of the copyright
 * holder.  Neither the name of the University nor the name of the
 * author may be used to endorse or promote products derived from this
 * material without specific prior written permission.  THIS SOFTWARE
 * IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR ANY PARTICULAR PURPOSE.
 *
 * <p>
 * 
 * Port of the BLOAT to use the Jakata Apache project BCEL 5.2 was 
 * contributed by The Australian National University by
 * @author Arrin Daley
 * @author John N Zigman
 * 
 * <p>
 * Java is a trademark of Sun Microsystems, Inc.
 */

package edu.purdue.cs.bloat.inline;

import java.util.*;

import org.apache.bcel.generic.*;

import edu.purdue.cs.bloat.cfg.FlowGraph;
import edu.purdue.cs.bloat.editor.InstructionAdapter;

/**
 * Used to keep track of the height of the stack. As instructions are visited,
 * the height of the stack is adjusted accordingly.
 */
public class StackHeightCounter extends InstructionAdapter {
    public static boolean DEBUG = false;

    private int height; // Current stack height

    private LinkedHashMap labelHeights; // Maps labels to their heights as

    // Integers

    private MethodGen method; // Method whose height we're computing

    Set tryCatches; // TryCatches active at current instruction

    private static void db(String s) {
        if (DEBUG)
            System.out.println(s);
    }

    /**
     * Constructor.
     * 
     * @param height
     *            Initial height of stack
     */
    public StackHeightCounter(MethodGen method) {
        this.method = method;
        this.height = 0;
        this.labelHeights = new LinkedHashMap();
        this.tryCatches = new LinkedHashSet();
    }

    /**
     * Returns the current height of the stack.
     */
    public int height() {
        return (this.height);
    }

    /**
     * Handles a Label. Special provisions must be made for labels that catch
     * exceptions.
     */
    // public void handle(Label label) {
    // }
    // EVERYTHING IS A INSTRUCTION NOW
    /**
     * Handles an instruction. Special provisions must be made to handle jumps,
     * switches, throws, and returns.
     */
    public void handle(InstructionHandle ih) {
        if (FlowGraph.label(ih)) {
            Integer labelHeight = (Integer) labelHeights.get(ih);
            if (labelHeight != null) {
                height = labelHeight.intValue();
            }

            // If this label begins an exception handler, then start it off
            // with a new stack with one element (the exception object) on
            // it.
            CodeExceptionGen[] tryCatches = method.getExceptionHandlers();
            for (int i = 0; i < tryCatches.length; i++) {
                if (tryCatches[i].getHandlerPC().equals(ih)) {
                    FlowGraph.setLabel(ih);
                    height = 1;
                    break;
                }

                if (tryCatches[i].getStartPC().equals(ih)) {
                    // If this block starts a protected region make note of
                    // the
                    // TryCatch block
                    this.tryCatches.add(tryCatches[i]);
                }

                if (tryCatches[i].getEndPC().equals(ih)) {
                    // If this block ends a protected region, remove it from
                    // the
                    // tryCatches list
                    this.tryCatches.remove(tryCatches[i]);
                }
            }
        }// ends the bit where we handle Labels but because a labels are
        // Instructions also....
        Instruction inst = ih.getInstruction();
        // inst.visit(this);
        // Hopefully this avoids that annoying switch statement and all
        // those visits
        int delta = 0;
        if (inst instanceof StackConsumer)
            delta = delta
                    - ((StackConsumer) inst).consumeStack(method
                            .getConstantPool());
        if ((inst instanceof StackProducer))// &&!(inst instanceof
            // JsrInstruction))
            delta = delta
                    - ((StackProducer) inst).produceStack(method
                            .getConstantPool());
        this.height = this.height + delta;

        if (inst instanceof IfInstruction || inst instanceof GotoInstruction) {
            BranchInstruction bI = (BranchInstruction) inst;
            InstructionHandle target = bI.getTarget();
            FlowGraph.setLabel(target);
            Integer targetHeight = (Integer) labelHeights.get(target);
            if (targetHeight != null) {
                if (targetHeight.intValue() != height) {
                    // Make sure stack heights match
                    db("Stack height mismatch (" + targetHeight.intValue()
                            + " != " + height + ") at " + inst);
                }

            } else {
                labelHeights.put(target, new Integer(height));
            }

        } else if (inst instanceof Select) {
            // Propagate height to all targets
            Select sw = (Select) inst;

            InstructionHandle[] targets = sw.getTargets();
            for (int t = 0; t < targets.length; t++) {
                InstructionHandle target = targets[t];
                FlowGraph.setLabel(target);
                Integer targetHeight = (Integer) labelHeights.get(target);
                if (targetHeight != null) {
                    if (targetHeight.intValue() != height) {
                        // Make sure stack heights match
                        db("Stack height mismatch (" + targetHeight.intValue()
                                + " != " + height + ") at " + inst);
                    }
                } else {
                    labelHeights.put(target, new Integer(height));
                }
            }

        } else if (inst instanceof JsrInstruction) {
            // We have to account for the return address being pushed on the
            // stack. Let's ignore the fact that someday in the future
            // subroutines may push stuff on the stack. M'kay?
            InstructionHandle subroutine = ((JsrInstruction) inst).getTarget();
            FlowGraph.setLabel(subroutine);
            Integer subHeight = (Integer) labelHeights.get(subroutine);
            if (subHeight != null) {
                if (subHeight.intValue() != height + 1) {
                    db("Stack height mismatch at subroutine ("
                            + subHeight.intValue() + " != " + (height + 1)
                            + ") at " + inst);
                }

            } else {
                labelHeights.put(subroutine, new Integer(height + 1));
            }

        } else if (inst instanceof ExceptionThrower
                || inst instanceof ReturnInstruction) {
            // Clear the stack
            height = 0;
        }
    }

    /**
     * Simulates the effect of "backing up" over an instruction.
     */
    public void unhandle(InstructionHandle inst) {
        // Temporarily negate the stack height, perform the normal handle,
        // and then negate the stack height again.
        this.height = -this.height;
        this.handle(inst);
        this.height = -this.height;
    }

    /**
     * Returns the set of <tt>TryCatch</tt> objects for the protected region
     * that the current instruction may be in.
     */
    public Set tryCatches() {
        return (this.tryCatches);
    }

    /*
     * public void visit_lushr(Instruction inst) { // Yes, it's only -1. The
     * long and the int index are popped off // and the shifted value is pushed.
     * Net loss of 1. height -= 1; }
     * 
     * public void visit_jsr(Instruction inst) { // Even though the jsr
     * instruction itself pushes the return // address onto the stack, we don't
     * want to account for that // here. It is already taken care of in the
     * handle method. This // way the label following the jsr (the return site)
     * will have the // stack height it had before the call. Once again, we do
     * not // account for the possibility of the jsr modifying the height of //
     * the stack. height += 0; }
     * 
     */
}
