/*
 * Class: Main   
 * Version: 
 * Date:
 * 
 * All files in the distribution of BLOAT (Bytecode Level Optimization and
 * Analysis tool for Java(tm)) are Copyright 1997-2001 by the Purdue
 * Research Foundation of Purdue University.  All rights reserved.
 *
 * <p>
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this entire copyright notice is duplicated in all
 * such copies, and that any documentation, announcements, and other
 * materials related to such distribution and use acknowledge that the
 * software was developed at Purdue University, West Lafayette, IN by
 * @author Antony Hosking
 * @author David Whitlock
 * @author Nathaniel Nystrom
 * No charge may be made for copies, derivations, or distributions of
 * this material without the express written consent of the copyright
 * holder.  Neither the name of the University nor the name of the
 * author may be used to endorse or promote products derived from this
 * material without specific prior written permission.  THIS SOFTWARE
 * IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR ANY PARTICULAR PURPOSE.
 *
 * <p>
 * 
 * Port of the BLOAT to use the Jakata Apache project BCEL 5.2 was 
 * contributed by The Australian National University by
 * @author Arrin Daley
 * @author John N Zigman
 * 
 * <p>
 * Java is a trademark of Sun Microsystems, Inc.
 */

package edu.purdue.cs.bloat.optimize;

import org.apache.bcel.util.ClassPath;
import org.apache.bcel.util.Repository;
import org.apache.bcel.generic.*;
import org.apache.bcel.classfile.*;

import edu.purdue.cs.bloat.cfg.DominatorTree;
import edu.purdue.cs.bloat.cfg.FlowGraph;
import edu.purdue.cs.bloat.cfg.VerifyCFG;
import edu.purdue.cs.bloat.codegen.CodeGenerator;
import edu.purdue.cs.bloat.codegen.Liveness;
import edu.purdue.cs.bloat.context.BloatContext;
import edu.purdue.cs.bloat.context.PersistentBloatContext;
import edu.purdue.cs.bloat.diva.InductionVarAnalyzer;
import edu.purdue.cs.bloat.editor.ClassHierarchy;
import edu.purdue.cs.bloat.inline.StackHeightCounter;
import edu.purdue.cs.bloat.ssa.SSA;
import edu.purdue.cs.bloat.ssa.SSAGraph;
import edu.purdue.cs.bloat.tbaa.TypeInference;
import edu.purdue.cs.bloat.trans.DeadCodeElimination;
import edu.purdue.cs.bloat.trans.ExprPropagation;
import edu.purdue.cs.bloat.trans.Peephole;
import edu.purdue.cs.bloat.trans.SSAPRE;
import edu.purdue.cs.bloat.trans.StackOpt;
import edu.purdue.cs.bloat.trans.StackOptimization;
import edu.purdue.cs.bloat.trans.ValueFolding;
import edu.purdue.cs.bloat.trans.ValueNumbering;
import edu.purdue.cs.bloat.tree.Tree;
import edu.purdue.cs.bloat.codegen.RegisterAllocator;

import java.io.*;
import java.text.*;
import java.util.*;

/**
 * Usage: java edu.purdue.cs.bloat.optimize.Main [-options] classes dir
 * 
 * where options include: -help print out this message -v -verbose turn on
 * verbose mode -debug display a hideous amount of debug info -classpath
 * <directories separated by colons> list directories in which to look for
 * classes -f optimize files even if up-to-date -closure recursively optimize
 * referenced classes -relax-loading don't report errors if a class is not found
 * -skip <class|package.*> skip the given class or package -only
 * <class|package.*> skip all but the given class or package -preserve-debug try
 * to preserve debug information -[no]anno insert an annotation in the contant
 * pool -[no]stack-alloc try to push locals onto the operand stack -peel-loops
 * <n|all> peel innermost loops to enable code hoisting (n >= 0 is the maximum
 * loop level to peel) -[no]pre perform partial redundency elimination
 * -[no]appre perform partial redundency elimination on access paths -[no]dce
 * perform dead code elimination -diva perform demand-driven induction variable
 * analysis -[no]prop perform copy and constant propagation
 * 
 */
public class Main {
    // the default class path to read the classes from
    static String DEFAULT_CLASSPATH = ".";

    // Flags that can be set/unset from the command line
    static boolean DEBUG = false; // Display debugging information

    static boolean VERBOSE = false;// Display status information as program

    // runs

    public static boolean TRACE = false;// Track our progress

    static boolean FORCE = false;// Optimize file even if they are

    // up-to-date

    static boolean CLOSURE = false;// Opimtize over entire heirarchy (i.e.

    // start

    // at the specified classes and recurse up the
    // class heirarchy)

    static DateFormat dateFormat = DateFormat.getDateTimeInstance(
            DateFormat.SHORT, DateFormat.LONG);

    static boolean DIVA = false;

    public static boolean PRE = true;

    public static boolean DCE = true;

    public static boolean PROP = true;

    public static boolean FOLD = true; // Value folding

    public static boolean INFER = true; // Type inference

    public static boolean NUMBER = true; // Value numbering

    public static boolean PERSIST = false; // Persistent check elimination

    public static boolean STACK_ALLOC = false;

    static boolean ANNO = true; // Note that class file was optimized

    public static boolean VERIFY = true;

    public static boolean OPT_STACK_1 = false; // Perform stack optimizations

    public static boolean OPT_STACK_2 = false;

    static String[] ARGS = null; // Arguments from the command line

    public static List SKIP = new ArrayList();

    // Classes that are specifically not optimized
    static List ONLY = new ArrayList();

    // Classes that are specifically optimized
    static String METHOD = null; // The name of one method to edit

    static BloatContext context = null;

    // static ClassFileLoader loader = null;//Used to load classes from
    // class files
    static Repository loader = null;

    static File outputDir = null;

    public static List optimizationsToPerform = null;

    /**
     * Parses the command line. The user must specify at least one class to
     * optimize and the directory in which to place the optimized class files.
     * The methods of the specified classes are then optimized according to the
     * command line options.
     * 
     * @see ClassEditor
     * @see ClassFileLoader
     * @see ClassFile
     * @see MethodEditor
     * @see MethodInfo
     * 
     * @see CompactArrayInitializer
     * @see FlowGraph
     * 
     */
    public static void main(String[] args) {
        try {
            ClassPath classPath = new ClassPath(DEFAULT_CLASSPATH);

            List classes = new ArrayList(args.length); // The classes to
            // optimize
            boolean gotdir = false; // Has an output directory been
            // specified?

            ARGS = args;

            for (int i = 0; i < args.length; i++) {
                if (args[i].equals("-v") || args[i].equals("-verbose")) {
                    VERBOSE = true;

                } else if (args[i].equals("-debug")) {
                    DEBUG = true;
                    FlowGraph.DEBUG = true;
                    DominatorTree.DEBUG = true;
                    Tree.DEBUG = true;
                    CodeGenerator.DEBUG = true;
                    Liveness.DEBUG = true;
                    SSA.DEBUG = true;
                    SSAGraph.DEBUG = true;
                    // PersistentCheckElimination.DEBUG = true;
                    ValueNumbering.DEBUG = true;
                    ValueFolding.DEBUG = true;
                    ClassHierarchy.DEBUG = true;
                    TypeInference.DEBUG = true;
                    SSAPRE.DEBUG = true;
                    ExprPropagation.DEBUG = true;
                    DeadCodeElimination.DEBUG = true;
                    CodeGenerator.DB_OPT_STACK = true;
                } else if (args[i].equals("-trace")) {
                    TRACE = true;
                } else if (args[i].equals("-db")) {

                    if (++i >= args.length) {
                        System.err.println("** No debugging option specified");
                        usage();
                    }

                    if (args[i].equals("bc")) {
                        // CodeArray.DEBUG = true;

                    } else if (args[i].equals("cfg")) {
                        FlowGraph.DEBUG = true;

                    } else if (args[i].equals("ssa")) {
                        SSA.DEBUG = true;
                        SSAGraph.DEBUG = true;

                    } else if (args[i].equals("graphs")) {
                        FlowGraph.DB_GRAPHS = true;

                    } else if (args[i].startsWith("-")) {
                        i--;

                    } else {
                        System.err.println("** Unknown debugging option: "
                                + args[i]);
                        usage();
                    }

                } else if (args[i].equals("-debugvf")) {
                    ValueFolding.DUMP = true;

                } else if (args[i].equals("-debugbc")) {
                    // BloatContext.DEBUG = true;

                } else if (args[i].equals("-help")) {
                    usage();

                } else if (args[i].equals("-noanno")) {
                    ANNO = false;

                } else if (args[i].equals("-anno")) {
                    ANNO = true;

                } else if (args[i].equals("-print-flow-graph")) {
                    FlowGraph.PRINT_GRAPH = true;

                } else if (args[i].equals("-preserve-debug")) {
                    // CodeGenerator.PRESERVE_DEBUG = true;

                } else if (args[i].equals("-nouse-stack-vars")) {
                    Tree.USE_STACK = false;

                } else if (args[i].equals("-use-stack-vars")) {
                    Tree.USE_STACK = true;

                } else if (args[i].equals("-unique-handlers")) {
                    // CodeGenerator.UNIQUE_HANDLERS = true;

                } else if (args[i].equals("-nostack-alloc")) {
                    STACK_ALLOC = false;

                } else if (args[i].equals("-stack-alloc")) {
                    STACK_ALLOC = true;

                } else if (args[i].equals("-no-verify")) {
                    VERIFY = false;

                } else if (args[i].equals("-peel-loops")) {
                    if (++i >= args.length) {
                        usage();
                    }

                    String n = args[i];

                    if (n.equals("all")) {
                        FlowGraph.PEEL_LOOPS_LEVEL = FlowGraph.PEEL_ALL_LOOPS;

                    } else {
                        try {
                            FlowGraph.PEEL_LOOPS_LEVEL = Integer.parseInt(n);

                            if (FlowGraph.PEEL_LOOPS_LEVEL < 0) {
                                usage();
                            }
                        } catch (NumberFormatException ex) {
                            usage();
                        }
                    }

                } else if (args[i].equals("-color")) {
                    Liveness.UNIQUE = false;

                } else if (args[i].equals("-nocolor")) {
                    Liveness.UNIQUE = true;

                } else if (args[i].equals("-only-method")) {
                    if (++i >= args.length) {
                        usage();
                    }

                    METHOD = args[i];

                } else if (args[i].equals("-classpath")) {
                    if (++i >= args.length) {
                        usage();
                    }

                    classPath = new ClassPath(args[i]);
                } else if (args[i].equals("-classpath/p")) {
                    if (++i >= args.length) {
                        usage();
                    }

                    classPath = new ClassPath(args[i]);
                } else if (args[i].equals("-skip")) {
                    if (++i >= args.length) {
                        usage();
                    }

                    String pkg = args[i];

                    // Account for class file name on command line
                    if (pkg.endsWith(".class"))
                        pkg = pkg.substring(0, pkg.lastIndexOf('.'));

                    SKIP.add(pkg.replace('.', '/'));

                } else if (args[i].equals("-only")) {
                    if (++i >= args.length) {
                        usage();
                    }

                    String pkg = args[i];

                    // Account for class file name on command line
                    if (pkg.endsWith(".class"))
                        pkg = pkg.substring(0, pkg.lastIndexOf('.'));

                    ONLY.add(pkg.replace('.', '/'));

                } else if (args[i].equals("-nodce")) {
                    DCE = false;

                } else if (args[i].equals("-noprop")) {
                    PROP = false;

                } else if (args[i].equals("-noappre")) {
                    SSAPRE.NO_ACCESS_PATHS = true;

                } else if (args[i].equals("-nopre")) {
                    PRE = false;

                } else if (args[i].equals("-dce")) {
                    DCE = true;

                } else if (args[i].equals("-prop")) {
                    PROP = true;

                } else if (args[i].equals("-appre")) {
                    SSAPRE.NO_ACCESS_PATHS = false;

                } else if (args[i].equals("-pre")) {
                    PRE = true;

                } else if (args[i].equals("-closure")) {
                    CLOSURE = true;

                } else if (args[i].equals("-opt-stack-1")) {
                    OPT_STACK_1 = true;
                    CodeGenerator.OPT_STACK = true;

                } else if (args[i].equals("-opt-stack-2")) {
                    OPT_STACK_2 = true;
                    CodeGenerator.OPT_STACK_2 = true;

                } else if (args[i].equals("-diva")) {
                    DIVA = true;

                } else if (args[i].equals("-no-thread")) {
                    SSAPRE.NO_THREAD = true;

                } else if (args[i].equals("-no-precise")) {
                    SSAPRE.NO_PRECISE = true;

                } else if (args[i].equals("-relax-loading")) {
                    ClassHierarchy.RELAX = true;

                } else if (args[i].equals("-f") || args[i].equals("-force")) {
                    FORCE = true;

                } else if (args[i].startsWith("-")) {
                    System.err.println("No such option: " + args[i]);
                    usage();

                } else if (i == args.length - 1) {
                    // Last argument is the name of the output directory

                    File f = new File(args[i]);

                    if (f.exists() && !f.isDirectory()) {
                        System.err.println("No such directory: " + f.getPath());
                        System.exit(2);
                    }

                    if (!f.exists()) {
                        f.mkdirs();
                    }

                    if (!f.exists()) {
                        System.err.println("Couldn't create directory: "
                                + f.getPath());
                        System.exit(2);
                    }

                    // Tell class loader to put optimized classes in f
                    // directory
                    outputDir = f;
                    gotdir = true;
                } else {
                    // The argument must be a class name...
                    classes.add(args[i]);
                }
            }

            loader = org.apache.bcel.util.SyntheticRepository
                    .getInstance(classPath);

            if (!gotdir) {
                System.err.println("No output directory specified");
                usage();
            }

            if (classes.size() == 0) {
                System.err.println("** No classes specified");
                usage();
            }

            if (TRACE)
                System.setProperty("edu.purdue.cs.bloat.TRACE", "true");
            if (DEBUG)
                System.setProperty("edu.purdue.cs.bloat.DEBUG", "true");

            // Use the CachingBloatingContext
            context = new PersistentBloatContext(loader, CLOSURE);

            buildOptimizationsToPerform();

            boolean errors = false;

            Iterator iter = classes.iterator();

            // Now that we've parsed the command line, load the classes into
            // the
            // class loader
            while (iter.hasNext()) {
                String name = (String) iter.next();

                try {
                    context.loadClass(name);

                } catch (ClassNotFoundException ex) {
                    System.err.println("Couldn't find class: "
                            + ex.getMessage());

                    errors = true;
                }
            }

            if (errors) {
                System.exit(1);
            }

            if (!CLOSURE) {
                Iterator e = classes.iterator();

                // Edit only the classes that were specified on the command line

                while (e.hasNext()) {
                    String name = (String) e.next();
                    editClass(name);
                }
            } else {
                // Edit all the classes in the class file editor and their
                // superclasses

                classes = null;

                if (TRACE) {
                    System.out.println("Computing closure "
                            + dateFormat.format(new Date()));
                }

                Iterator e = context.getHierarchy().classes().iterator();

                while (e.hasNext()) {
                    Type t = (Type) e.next();

                    if (t instanceof ObjectType) {
                        editClass(((ObjectType) t).getClassName());
                    }
                }
            }
        } catch (ExceptionInInitializerError ex) {
            ex.printStackTrace();
            System.out.println(ex.getException());
        }
    }

    private static void usage() {
        System.err
                .println("\nUsage: java edu.purdue.cs.bloat.optimize.Main"
                        + "\n            [-options] classes dir"
                        + "\n"
                        + "\nwhere options include:"
                        + "\n    -help             print out this message"
                        + "\n    -v -verbose       turn on verbose mode"
                        + "\n    -debug            display a hideous amount of debug info"
                        + "\n    -classpath <directories separated by colons>"
                        + "\n                      list directories in which to look for classes"
                        + "\n    -f                optimize files even if up-to-date"
                        + "\n    -closure          recursively optimize referenced classes"
                        + "\n    -relax-loading    don't report errors if a class is not found"
                        + "\n    -skip <class|package.*>"
                        + "\n                      skip the given class or package"
                        + "\n    -only <class|package.*>"
                        + "\n                      skip all but the given class or package"
                        + "\n    -preserve-debug   try to preserve debug information"
                        + "\n    -[no]anno         insert an annotation in the contant pool"
                        + "\n    -[no]stack-alloc  try to push locals onto the operand stack"
                        + "\n    -peel-loops <n|all>"
                        + "\n                      peel innermost loops to enable code hoisting"
                        + "\n                      (n >= 0 is the maximum loop level to peel)"
                        + "\n    -[no]pre          perform partial redundency elimination"
                        + "\n    -[no]dce          perform dead code elimination"
                        + "\n    -diva             perform demand-driven induction variable analysis"
                        + "\n    -[no]prop         perform copy and constant propagation"
                        + "");
        System.exit(0);
    }

    /**
     * Performs the actual editing of a class. Does a whole mess of stuff
     * including reading in the classfile, building data structures to represent
     * the class file, converting the CFG for each method in the class into SSA
     * form, perform some anlayses and optimizations on the method, and finally
     * committing it back to the class file. Phew.
     */
    private static void editClass(String className) {
        JavaClass classFile; // Holds info about a class (implements
        // ClassInfo)
        File targetFile = null;
        // Get information about the class className
        try {
            classFile = (JavaClass) context.loadClass(className);
        } catch (ClassNotFoundException ex) {
            System.err.println("** Couldn't find class: " + ex.getMessage());
            return;
        }

        // Check to see if the file is up-to-date (i.e. has been
        // recompiled since it was last optimized). If so, do nothing
        // because the FORCE flag is false.
        String source = classFile.getSourceFileName();
        System.out.println(source + " source file for class:"
                + classFile.getClassName());
        File sourceFile = new File(source);

        String target = outputDir.getAbsolutePath()
                + File.separatorChar
                + classFile.getClassName().substring(
                        classFile.getClassName().lastIndexOf(".") + 1)
                + ".class";
        System.out.println(target + " selected for output of class: "
                + classFile.getClassName());
        targetFile = new File(target);

        if (!FORCE) {
            if (sourceFile != null && targetFile != null && sourceFile.exists()
                    && targetFile.exists()
                    && sourceFile.lastModified() < targetFile.lastModified()) {
                if (VERBOSE) {
                    System.out.println(classFile.getClassName()
                            + " is up to date");
                }

                return;
            }
            // return;
        }

        if (DEBUG) {
            // Print the contents of the class file to System.out
            System.out.println(classFile);
        }

        ClassGen c = context.editClass(classFile);

        boolean skip = false;

        String name = c.getClassName();
        String qual = "";
        if (name.lastIndexOf('/') != -1)
            qual = name.substring(0, name.lastIndexOf('/')) + "/*";

        // Edit only classes explicitly mentioned.
        if (ONLY.size() > 0) {
            skip = true;

            // Only edit classes we explicitly don't name.
            for (int i = 0; i < ONLY.size(); i++) {
                String pkg = (String) ONLY.get(i);

                if (name.equals(pkg) || qual.equals(pkg)) {
                    skip = false;
                    break;
                }
            }
        }

        // Don't edit classes we explicitly skip.
        if (!skip) {
            for (int i = 0; i < SKIP.size(); i++) {
                String pkg = (String) SKIP.get(i);

                if (name.equals(pkg) || qual.equals(pkg)) {
                    skip = true;
                    break;
                }
            }
        }

        if (skip) {
            if (VERBOSE) {
                System.out.println("Skipping " + c.getClassName());
            }

            // We're done with this class file, decrement its reference
            // count
            context.release(classFile);
            return;
        }

        // Touch the output file first. That is, create the file, but make
        // it empty, just to make sure we can create it.
        // File outputDir = loader.outputDir();

        if (VERBOSE) {
            System.out.println("Optimizing " + c.getClassName());
        }

        // Finally, we can start playing with the methods...
        Method[] methods = c.getMethods();

        int numMethods = methods.length + 1;
        ;
        int whichMethod = 0;

        for (int j = 0; j < methods.length; j++) {
            final MethodGen m;

            try {
                m = context.editMethod(methods[j], c.getClassName());
            } catch (ClassFormatException ex) {
                System.err.println(ex.getMessage());
                continue;
            }

            if (TRACE) {
                whichMethod++;
                System.out
                        .println("Optimizing " + name + "." + m.getName()
                                + " (method " + whichMethod + " of "
                                + numMethods + ")");
            }

            if (METHOD != null) {
                // A method name has been specified on the command line using
                // -only-method.
                boolean pass = true;

                String t = m.getName() + m.getSignature();

                if (t.equals(METHOD)) {
                    pass = false;
                }

                t = m.getName();

                if (t.equals(METHOD)) {
                    pass = false;
                }

                if (pass) {
                    // This isn't the method we're looking for.
                    // Decrement its reference count.
                    context.release(methods[j], c.getClassName());
                    continue;
                }
            }

            if (DEBUG) {
                System.out.println(m);
            }

            if (m.isNative() || m.isAbstract()) {
                // We can't edit native or abstract methods
                context.release(methods[j], c.getClassName());
                continue;
            }

            MethodBloater bloater = new MethodBloater(optimizationsToPerform);
            bloater.bloatMethod(m, context);
        }

        if (ANNO) {
            String s = "Optimized with: edu.purdue.cs.bloat.optimize.Main";

            for (int i = 0; i < ARGS.length; i++) {
                if (ARGS[i].indexOf(' ') >= 0 || ARGS[i].indexOf('\t') >= 0
                        || ARGS[i].indexOf('\r') >= 0
                        || ARGS[i].indexOf('\n') >= 0) {
                    s += " '" + ARGS[i] + "'";
                } else {
                    s += " " + ARGS[i];
                }
            }

            System.out.println(s);
        }

        try {
            classFile = context.editClass(classFile.getClassName())
                    .getJavaClass();
        } catch (ClassNotFoundException cnfe) {
        }

        try {
            classFile.dump(targetFile);
        } catch (java.io.IOException io) {
            System.out.println("An error occured whilst writing file: "
                    + targetFile.getPath());
        }
        if (TRACE)
            System.out.println(context.toString());
    }

    public static void dumpcode(MethodGen m) {
        PrintWriter out = new PrintWriter(System.out, true);
        StackHeightCounter shc = new StackHeightCounter(m);

        out.println("Code for method " + m.getName() + m.getSignature());
        InstructionList instructions = m.getInstructionList();
        Iterator iter = instructions.iterator();
        while (iter.hasNext()) {
            Object obj = iter.next();
            shc.handle((InstructionHandle) obj);

            System.out
                    .println("        " + obj + " (sh: " + shc.height() + ")");
        }
    }

    /*
     * The following block of code is a version of buildOptimizationsToPerform
     * that does not use OptimizationsBuilder. If you revert to this version,
     * you will have to uncomment the last couple of lines of
     * MethodBloater#doAnOptimization().
     * 
     * private static void buildOptimizationsToPerform() {
     * optimizationsToPerform = new ArrayList();
     * 
     * if (COMPACT_ARRAY_INIT) optimizationsToPerform.add(new
     * CompactArrayInitializer());
     * 
     * optimizationsToPerform.add(new SSA());
     * 
     * if ((! Tree.USE_STACK) && PROP) optimizationsToPerform.add(new
     * ExprPropagation());
     * 
     * if (DCE) optimizationsToPerform.add(new DeadCodeElimination());
     * 
     * if (INFER) optimizationsToPerform.add(new TypeInference());
     * 
     * if (NUMBER) optimizationsToPerform.add(new ValueNumbering());
     * 
     * if (FOLD) optimizationsToPerform.add(new ValueFolding());
     * 
     * if (PRE) optimizationsToPerform.add(new SSAPRE());
     * 
     * if (FOLD) optimizationsToPerform.add(new ValueFolding());
     * 
     * if (PROP) optimizationsToPerform.add(new ExprPropagation()); // make sure
     * we've done at least one thing since the last DCE if (DCE && (INFER ||
     * NUMBER || FOLD || PRE || PROP)) optimizationsToPerform.add(new
     * DeadCodeElimination()); //doDeadCodeElimination(cfg, context); /* if
     * (PERSIST) (new PersistentCheckElimination()).transform(cfg);
     */
    /*
     * if (DIVA) optimizationsToPerform.add(new InductionVarAnalyzer());
     * //doInductionVarAnalysis(cfg, context);
     * 
     * if (OPT_STACK_2) { optimizationsToPerform.add(new StackOpt());
     * optimizationsToPerform.add(new SSA()); optimizationsToPerform.add(new
     * DeadCodeElimination()); }
     * 
     * if (CodeGenerator.OPT_STACK) { optimizationsToPerform.add(new
     * StackOptimization()); }
     * 
     * if (VERIFY) optimizationsToPerform.add(new VerifyCFG());
     * 
     * optimizationsToPerform.add(new CodeGen()); optimizationsToPerform.add(new
     * ConstantPoolSizePrinter()); optimizationsToPerform.add(new Peephole()); }
     */

    public static void buildOptimizationsToPerform() {

        OptimizationsBuilder builder = new OptimizationsBuilder();

        builder.add(new SSA());

        if ((!Tree.USE_STACK) && PROP)
            builder.add(new ExprPropagation());

        if (DCE)
            builder.add(new DeadCodeElimination());

        if (INFER)
            builder.add(new TypeInference());

        if (NUMBER)
            builder.add(new ValueNumbering());

        if (FOLD)
            builder.add(new ValueFolding());

        if (PRE)
            builder.add(new SSAPRE());

        if (FOLD)
            builder.add(new ValueFolding());

        if (PROP)
            builder.add(new ExprPropagation());

        // make sure we've done at least one thing since the last DCE
        if (DCE && (INFER || NUMBER || FOLD || PRE || PROP))
            builder.add(new DeadCodeElimination());

        /*
         * if (PERSIST) (new PersistentCheckElimination()).transform(cfg);
         */

        if (DIVA)
            builder.add(new InductionVarAnalyzer());

        if (OPT_STACK_2) {
            builder.add(new StackOpt());
            builder.add(new SSA());
            builder.add(new DeadCodeElimination());
        }
        RegisterAllocator alloc = new RegisterAllocator();
        builder.add(alloc);

        if (CodeGenerator.OPT_STACK) {
            builder.add(new StackOptimization());
        }

        if (VERIFY)
            builder.add(new VerifyCFG());

        builder.add(new CodeGen(alloc));
        // builder.add(new ConstantPoolSizePrinter());
        builder.add(new Peephole());

        optimizationsToPerform = builder.getList();
    }

    /**
     * OptimizationsBuilder is a little utility class intended to simplify the
     * creation of the list of optimizations. In particular, it will
     * automatically add a VerifyCFG call after each optimization whenever we're
     * in debug mode. This simplifies the creation of the list, and
     * {@link MethodBloater#doAnOptimization(Optimization, MethodState)} too.
     * 
     * @author Chris Bennetts
     * 
     */
    static class OptimizationsBuilder {
        private List _optimizations = new ArrayList();

        public List getList() {
            return _optimizations;
        }

        public OptimizationsBuilder add(Optimization opt) {
            _optimizations.add(opt);
            if (DEBUG)
                _optimizations.add(new VerifyCFG());

            return this;
        }
    }
}
