/*
 * Class: MethodBloater   
 * Version: 
 * Date:
 * 
 * All files in the distribution of BLOAT (Bytecode Level Optimization and
 * Analysis tool for Java(tm)) are Copyright 1997-2001 by the Purdue
 * Research Foundation of Purdue University.  All rights reserved.
 *
 * <p>
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this entire copyright notice is duplicated in all
 * such copies, and that any documentation, announcements, and other
 * materials related to such distribution and use acknowledge that the
 * software was developed at Purdue University, West Lafayette, IN by
 * @author Antony Hosking
 * @author David Whitlock
 * @author Nathaniel Nystrom
 * No charge may be made for copies, derivations, or distributions of
 * this material without the express written consent of the copyright
 * holder.  Neither the name of the University nor the name of the
 * author may be used to endorse or promote products derived from this
 * material without specific prior written permission.  THIS SOFTWARE
 * IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR ANY PARTICULAR PURPOSE.
 *
 * <p>
 * 
 * Port of the BLOAT to use the Jakata Apache project BCEL 5.2 was 
 * contributed by The Australian National University by
 * @author Arrin Daley
 * @author John N Zigman
 * 
 * <p>
 * Java is a trademark of Sun Microsystems, Inc.
 */

package edu.purdue.cs.bloat.optimize;

import java.text.DateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import org.apache.bcel.generic.LocalVariableGen;
import org.apache.bcel.generic.CodeExceptionGen;
import org.apache.bcel.generic.MethodGen;

import edu.purdue.cs.bloat.context.BloatContext;
import edu.purdue.cs.bloat.util.Assert;

/**
 * Performs BLOAT-ing on a method. A single MethodBloater object may be safely
 * reused for multiple methods.
 * 
 * Extra debugging information will be printed if the system property
 * <tt>edu.purdue.cs.bloat.DEBUG</tt> is set to <tt>true</tt>.
 * 
 * @author Chris Bennetts
 * 
 */
public class MethodBloater {
    static DateFormat dateFormat = DateFormat.getDateTimeInstance(
            DateFormat.SHORT, DateFormat.LONG);

    static boolean DEBUG = "true".equals(System
            .getProperty("edu.purdue.cs.bloat.DEBUG"));

    List optimizationsToPerform;

    /**
     * Create a new MethodBloater.
     * 
     * @param optimizationsToPerform
     *            The optimizations to be performed.
     */
    public MethodBloater(List optimizationsToPerform) {
        this.optimizationsToPerform = optimizationsToPerform;
    }

    /**
     * BLOAT a single method. Creates the {@link MethodState method state},
     * runs the optimizations, and commits any changes back to the context
     * object.
     * <p>
     * 
     * Will return prematurely if there is a problem creating the control flow
     * graph.
     */
    public void bloatMethod(MethodGen methodGen, BloatContext context) {
        try {
            MethodBloater.printMethodGenInfo(methodGen);

            MethodState state = new MethodState(methodGen, context);
            if (state.controlFlowGraph() == null)
                return;

            performOptimizations(state);
            state.commitChanges();

            // ConstantPoolSizePrinter.print(state);
            MethodBloater.printMethodGenInfo(methodGen);
        } catch (IllegalArgumentException e) {
            System.out.println(" NOTE: CFG did not verify while bloating "
                    + methodGen.getName()
                    + " after all optimizations. Exception: " + e);
            e.printStackTrace();
        } catch (Exception e) {
            String msg = "** Exception while optimizing " + methodGen.getName()
                    + methodGen.getSignature() + " of class "
                    + methodGen.getClassName();
            System.err.println(msg);
            System.err.println(e.getMessage());
            e.printStackTrace(System.err);
            System.exit(1);
        }
    }

    /**
     * Performs all optimizations on the method.
     * 
     * @param state
     *            The method to perform the optimizations on.
     * @throws Exception
     *             Any exception that may be thrown by the optimizations.
     */
    public void performOptimizations(MethodState state) throws Exception {
        Assert.isNotNull(optimizationsToPerform);

        Iterator it = optimizationsToPerform.iterator();

        while (it.hasNext()) {
            Optimization opt = (Optimization) it.next();
            doAnOptimization(opt, state);
        }
    }

    /**
     * Print the trace message for an optimization.
     * 
     * @see Optimization#traceMessage(String)
     */
    private void printTraceMessage(Optimization o) {
        if (!DEBUG)
            return;

        String dateString = dateFormat.format(new Date());
        String message = o.traceMessage(dateString);

        if (message != null)
            System.out.println(message);
    }

    /**
     * Print a debug message for an optimization. Called before the
     * transformation is performed.
     * 
     * @see Optimization#preDebugMessage()
     */
    private void printPreDebugMessage(Optimization o) {
        if (!DEBUG)
            return;

        String message = o.preDebugMessage();

        if (message != null)
            System.out.println(message);
    }

    /**
     * Print a debug message for an optimization. Called after the
     * transformation is performed.
     * 
     * @see Optimization#postDebugMessage()
     */
    private void printPostDebugMessage(Optimization o) {
        if (!DEBUG)
            return;

        String message = o.postDebugMessage();

        if (message != null)
            System.out.println(message);
    }

    /**
     * Apply an optimization to a method.
     * 
     * @param optimization
     *            The optimization to be performed.
     * @param state
     *            The method to perform the optimization on.
     * @throws Exception
     *             Any exception thrown by the optimization.
     * 
     * @see Optimization#transform(MethodState)
     */
    private void doAnOptimization(Optimization optimization, MethodState state)
            throws Exception {
        printTraceMessage(optimization);
        printPreDebugMessage(optimization);

        optimization.transform(state);

        printPostDebugMessage(optimization);

        // FIXME: VerifyCFG can itself now be called as an Optimization.
        // Use a Builder or something to encapsulate this bit of behaviour.
        // This is currently implemented by Main.OptimizationsBuilder, and
        // MainG2 allows it to just be inserted into the list of optimizations.

        // if (DEBUG)
        // VerifyCFG.verify(state.controlFlowGraph());
    }

    /**
     * Print some debugging info about the MethodGen.
     */
    static void printMethodGenInfo(MethodGen methodGen) {
        if (!DEBUG)
            return;

        LocalVariableGen[] locals = null;
        CodeExceptionGen[] excptns = null;
        System.out.println("Method: " + methodGen);
        System.out.println(methodGen.getMethod().getCode());
        System.out.println();
        locals = methodGen.getLocalVariables();
        if (locals != null) {
            System.out.println("LocalVariables: ");
            for (int j = 0; j < locals.length; j++)
                System.out.println("  " + locals[j]);
        }
        excptns = methodGen.getExceptionHandlers();
        if (excptns != null) {
            System.out.println("ExceptionHandlers: ");
            for (int j = 0; j < excptns.length; j++)
                System.out.println("  " + excptns[j]);
        }
    }
}
