/*
 * Class: MethodState   
 * Version: 
 * Date:
 * 
 * All files in the distribution of BLOAT (Bytecode Level Optimization and
 * Analysis tool for Java(tm)) are Copyright 1997-2001 by the Purdue
 * Research Foundation of Purdue University.  All rights reserved.
 *
 * <p>
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this entire copyright notice is duplicated in all
 * such copies, and that any documentation, announcements, and other
 * materials related to such distribution and use acknowledge that the
 * software was developed at Purdue University, West Lafayette, IN by
 * @author Antony Hosking
 * @author David Whitlock
 * @author Nathaniel Nystrom
 * No charge may be made for copies, derivations, or distributions of
 * this material without the express written consent of the copyright
 * holder.  Neither the name of the University nor the name of the
 * author may be used to endorse or promote products derived from this
 * material without specific prior written permission.  THIS SOFTWARE
 * IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR ANY PARTICULAR PURPOSE.
 *
 * <p>
 * 
 * Port of the BLOAT to use the Jakata Apache project BCEL 5.2 was 
 * contributed by The Australian National University by
 * @author Arrin Daley
 * @author John N Zigman
 * 
 * <p>
 * Java is a trademark of Sun Microsystems, Inc.
 */

package edu.purdue.cs.bloat.optimize;

import org.apache.bcel.classfile.ClassFormatException;
import org.apache.bcel.generic.MethodGen;

import edu.purdue.cs.bloat.cfg.FlowGraph;
import edu.purdue.cs.bloat.context.BloatContext;

/**
 * MethodState encapsulates three related items for a method:
 * 
 * <ul>
 * <li>its MethodGen,</li>
 * <li>its Control Flow Graph,</li>
 * <li>its context.</li>
 * </ul>
 * 
 * These tend to be passed around and updated together, so it made sense to
 * group them together.
 * 
 * Note that this class has no &quot;setters&quot;. If they appear necessary,
 * move the code that would need them in to this class, instead of taking the
 * shortcut.
 * 
 * @author Chris Bennetts
 * 
 */
public class MethodState {
    private MethodGen _methodGen;

    private FlowGraph _controlFlowGraph;

    private BloatContext _context;

    /**
     * Create a new MethodState object. The control flow graph is generated
     * automatically.
     */
    public MethodState(MethodGen methodGen, BloatContext context) {
        _methodGen = methodGen;
        _context = context;

        rebuildFlowGraph();
    }

    /**
     * Obtain the MethodGen.
     */
    public MethodGen methodGen() {
        return _methodGen;
    }

    /**
     * Obtain the control flow graph. In general, favour
     * {@link #controlFlowGraph() controlFlowGraph()}.
     */
    public FlowGraph cfg() {
        return controlFlowGraph();
    }

    /**
     * Obtain the control flow graph.
     */
    public FlowGraph controlFlowGraph() {
        return _controlFlowGraph;
    }

    /**
     * Obtain the current context.
     */
    public BloatContext context() {
        return _context;
    }

    /**
     * Rebuild the control flow graph. Use if it falls out of sync with the
     * MethodGen.
     * 
     */
    public void rebuildFlowGraph() {
        try {
            _controlFlowGraph = new FlowGraph(_methodGen);
        } catch (ClassFormatException ex) {
            System.err.println(ex.getMessage());
            _context.release(_methodGen.getMethod(), _methodGen.getClassName());
            _controlFlowGraph = null;

            return;
        }

        _controlFlowGraph.initialize();
    }

    /**
     * Commit changes to the method to the context.
     * 
     */
    public void commitChanges() {
        _context.commit(_methodGen);
    }
}
