/*
 * Class: TBAA   
 * Version: 
 * Date:
 * 
 * All files in the distribution of BLOAT (Bytecode Level Optimization and
 * Analysis tool for Java(tm)) are Copyright 1997-2001 by the Purdue
 * Research Foundation of Purdue University.  All rights reserved.
 *
 * <p>
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this entire copyright notice is duplicated in all
 * such copies, and that any documentation, announcements, and other
 * materials related to such distribution and use acknowledge that the
 * software was developed at Purdue University, West Lafayette, IN by
 * @author Antony Hosking
 * @author David Whitlock
 * @author Nathaniel Nystrom
 * No charge may be made for copies, derivations, or distributions of
 * this material without the express written consent of the copyright
 * holder.  Neither the name of the University nor the name of the
 * author may be used to endorse or promote products derived from this
 * material without specific prior written permission.  THIS SOFTWARE
 * IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR ANY PARTICULAR PURPOSE.
 *
 * <p>
 * 
 * Port of the BLOAT to use the Jakata Apache project BCEL 5.2 was 
 * contributed by The Australian National University by
 * @author Arrin Daley
 * @author John N Zigman
 * 
 * <p>
 * Java is a trademark of Sun Microsystems, Inc.
 */

package edu.purdue.cs.bloat.tbaa;

import org.apache.bcel.generic.*;
import org.apache.bcel.classfile.*;

import edu.purdue.cs.bloat.editor.ClassHierarchy;
import edu.purdue.cs.bloat.editor.EditorContext;
import edu.purdue.cs.bloat.editor.MemberRef;
import edu.purdue.cs.bloat.tree.ArrayRefExpr;
import edu.purdue.cs.bloat.tree.ConstantExpr;
import edu.purdue.cs.bloat.tree.Expr;
import edu.purdue.cs.bloat.tree.FieldExpr;
import edu.purdue.cs.bloat.tree.MemRefExpr;
import edu.purdue.cs.bloat.tree.StaticFieldExpr;
import edu.purdue.cs.bloat.util.Assert;

import java.util.*;

/**
 * Performs Type-Based Alias Analysis (TBAA) to determine if one expression can
 * alias another. An expression may alias another expression if there is the
 * possibility that they refer to the same location in memory. BLOAT models
 * expressions that may reference memory locations with <tt>MemRefExpr</tt>s.
 * There are two kinds of "access expressions" in Java: field accesses (<tt>FieldExpr</tt>
 * and <tt>StaticFieldExpr</tt>) and array references (<tt>ArrayRefExpr</tt>).
 * Access paths consist of one or more access expressions. For instance,
 * <tt>a.b[i].c</tt> is an access expression.
 * 
 * <p>
 * 
 * TBAA uses the FieldTypeDecl relation to determine whether or not two access
 * paths may refer to the same memory location.
 * 
 * <p>
 * 
 * <table>
 * <th>
 * <td>AP1 </td>
 * <td>AP2 </td>
 * <td>FieldTypeDecl(AP1, Ap2)</td>
 * </th>
 * <tr>
 * <Td>p </td>
 * <Td>p </td>
 * <td>true </td>
 * </tr>
 * <tr>
 * <td>p.f </td>
 * <td>q.g </td>
 * <td>(f = g) && FieldTypeDecl(p, q)</td>
 * </tr>
 * <tr>
 * <td>p.f </td>
 * <Td>q[i]</td>
 * <td>false </td>
 * </tr>
 * <tr>
 * <td>p[i]</td>
 * <td>q[j]</td>
 * <td>FieldTypeDecl(p, q) </td>
 * </tr>
 * <tr>
 * <td>p </td>
 * <td>q </td>
 * <td>TypeDecl(p, q) </td>
 * </tr>
 * </table>
 * 
 * <p>
 * 
 * The TypeDecl(AP1, AP2) relation is defined as:
 * <p align=center>
 * Subtypes(Type(AP1)) INTERSECT Subtypes(Type(AP2)) != EMPTY
 * </p>
 * 
 * The subtype relationships are determined by the <tt>ClassHierarchy</tt>.
 * 
 * @see ClassHierarchy
 * @see MemRefExpr
 * @see FieldExpr
 * @see StaticFieldExpr
 * @see ArrayRefExpr
 */
public class TBAA {

    /**
     * Returns true, if expression <tt>a</tt> can alias expression <tt>b</tt>.
     */
    public static boolean canAlias(EditorContext context, Expr a, Expr b) {
        // Only memory reference expressions can have aliases.
        if (!(a instanceof MemRefExpr)) {
            return false;
        }

        // Only memory reference expressions can have aliases.
        if (!(b instanceof MemRefExpr)) {
            return false;
        }

        // Equal expressions can be aliases.
        if (a.equalsExpr(b)) {
            return true;
        }

        MemberRef af = null; // Field accessed by expression a
        MemberRef bf = null; // Field accessed by expression b

        if (a instanceof FieldExpr) {
            af = ((FieldExpr) a).field();
        }

        if (a instanceof StaticFieldExpr) {
            af = ((StaticFieldExpr) a).field();
        }

        if (b instanceof FieldExpr) {
            bf = ((FieldExpr) b).field();
        }

        if (b instanceof StaticFieldExpr) {
            bf = ((StaticFieldExpr) b).field();
        }

        // Arrays and fields cannot alias the same location.
        if (a instanceof ArrayRefExpr && bf != null) {
            return false;
        }

        // Arrays and fields cannot alias the same location.
        if (b instanceof ArrayRefExpr && af != null) {
            return false;
        }

        Collection types = new LinkedHashSet();
        types.add(a.type());
        types.add(b.type());
        ClassHierarchy hier = new ClassHierarchy(context, types, true);

        // Only type-compatible arrays can alias the same location.
        if (a instanceof ArrayRefExpr && b instanceof ArrayRefExpr) {
            ArrayRefExpr aa = (ArrayRefExpr) a;
            ArrayRefExpr bb = (ArrayRefExpr) b;

            Type aaIndexType = aa.index().type();
            Type bbIndexType = bb.index().type();
            Type aaArrayType = aa.array().type();
            Type bbArrayType = bb.array().type();

            Assert.isTrue(Assert.isIntegral(aaIndexType), aa.index() + " in "
                    + aa + " (" + aaIndexType + ") is not an integer");
            Assert.isTrue(Assert.isIntegral(bbIndexType), bb.index() + " in "
                    + bb + " (" + bbIndexType + ") is not an integer");
            Assert.isTrue(aaArrayType instanceof ArrayType
                    || aaArrayType.equals(Type.OBJECT)
                    || aaArrayType.equals(new ObjectType(
                            "java.lang.Serializable"))
                    || aaArrayType.equals(new ObjectType("jav.lang.Cloneable"))
                    || aaArrayType.equals(Type.NULL), aa.array() + " in " + aa
                    + " (" + aaArrayType + ") is not an array");
            Assert.isTrue(bbArrayType instanceof ArrayType
                    || bbArrayType.equals(Type.OBJECT)
                    || bbArrayType.equals(new ObjectType(
                            "java.lang.Serializable"))
                    || bbArrayType.equals(new ObjectType("jav.lang.Cloneable"))
                    || bbArrayType.equals(Type.NULL), bb.array() + " in " + bb
                    + " (" + bbArrayType + ") is not an array");

            // Optimization: if constant indices. Only equal indices can
            // alias the same location.
            if (aa.index() instanceof ConstantExpr
                    && bb.index() instanceof ConstantExpr) {

                ConstantExpr ai = (ConstantExpr) aa.index();
                ConstantExpr bi = (ConstantExpr) bb.index();

                if (ai.value() != null && bi.value() != null) {
                    if (!ai.value().equals(bi.value())) {
                        return false;
                    }
                }
            }

            return intersects(hier, aaArrayType, bbArrayType);
        }

        try {
            if (af != null) {
                ClassGen klass = context.editClass(((ObjectType) af
                        .declaringClass()).getClassName());
                Field e = klass.containsField(af.name());
                if (e == null) {
                    throw new NoSuchFieldException("Field: " + af.name()
                            + "not found in class " + af.declaringClass());
                }

                if (e.isVolatile()) {
                    // context.release(e.fieldInfo());
                    return true;
                }

                if (e.isFinal()) {
                    // context.release(e.fieldInfo());
                    return false;
                }

                // context.release(e.fieldInfo());
                // WE NEVER STARTED EDITING SO DON'T HAVE TO RELEASE IT
            }

            if (bf != null) {
                ClassGen klass = context.editClass(((ObjectType) bf
                        .declaringClass()).getClassName());
                Field e = klass.containsField(bf.name());
                if (e == null) {
                    throw new NoSuchFieldException("Field: " + bf.name()
                            + "not found in class " + bf.declaringClass());
                }

                if (e.isVolatile()) {
                    // context.release(e.fieldInfo());
                    return true;
                }

                if (e.isFinal()) {
                    // context.release(e.fieldInfo());
                    return false;
                }

                // context.release(e.fieldInfo());
                // WE NEVER STARTED EDITING SO DON'T HAVE TO RELEASE IT
            }

        } catch (NoSuchFieldException e) {
            // A field wasn't found. Silently assume there is an alias.
            return true;
        } catch (ClassNotFoundException cnfe) {
            // A field wasn't found. Silently assume there is an alias.
            return true;
        }

        // Only fields with the same name can alias the same location.
        if (af != null && bf != null) {
            return af.equals(bf);
        }

        // Default case. This shouldn't happen.
        return intersects(hier, a.type(), b.type());
    }

    /**
     * Returns <tt>true</tt> if type a and type c intersect. That is, the two
     * types have a non-null intersection.
     */
    private static boolean intersects(ClassHierarchy hier, Type a, Type b) {
        return !hier.intersectType(a, b).equals(Type.NULL);
    }
}
