/*
 * Class: Tree    
 * Version: 
 * Date:
 * 
 * All files in the distribution of BLOAT (Bytecode Level Optimization and
 * Analysis tool for Java(tm)) are Copyright 1997-2001 by the Purdue
 * Research Foundation of Purdue University.  All rights reserved.
 *
 * <p>
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this entire copyright notice is duplicated in all
 * such copies, and that any documentation, announcements, and other
 * materials related to such distribution and use acknowledge that the
 * software was developed at Purdue University, West Lafayette, IN by
 * @author Antony Hosking
 * @author David Whitlock
 * @author Nathaniel Nystrom
 * No charge may be made for copies, derivations, or distributions of
 * this material without the express written consent of the copyright
 * holder.  Neither the name of the University nor the name of the
 * author may be used to endorse or promote products derived from this
 * material without specific prior written permission.  THIS SOFTWARE
 * IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR ANY PARTICULAR PURPOSE.
 *
 * <p>
 * 
 * Port of the BLOAT to use the Jakata Apache project BCEL 5.2 was 
 * contributed by The Australian National University by
 * @author Arrin Daley
 * @author John N Zigman
 * 
 * <p>
 * Java is a trademark of Sun Microsystems, Inc.
 */

package edu.purdue.cs.bloat.tree;

import java.util.*;
import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.apache.bcel.classfile.ClassFormatException;

import edu.purdue.cs.bloat.cfg.Block;
import edu.purdue.cs.bloat.cfg.Subroutine;
import edu.purdue.cs.bloat.editor.InstructionVisitor;
import edu.purdue.cs.bloat.editor.MemberRef;
import edu.purdue.cs.bloat.editor.MethodRef;
import edu.purdue.cs.bloat.editor.NameAndType;
import edu.purdue.cs.bloat.util.Assert;

/**
 * Tree represents the expression tree of a basic Block. It consists of an
 * operand (expression) stack comprised of expressions and a list of statements
 * contained in the block.
 * 
 * @see Block
 * @see Expr
 * @see OperandStack
 * @see Stmt see StmtList
 */
public class Tree extends Node implements InstructionVisitor,
        org.apache.bcel.generic.Visitor {
    public static boolean DEBUG = false;

    public static boolean FLATTEN = false;

    public static boolean USE_STACK = false; // Do we use stack vars?

    public static boolean AUPDATE_FIX_HACK = false;

    public static boolean AUPDATE_FIX_HACK_CHANGED = false;

    public static boolean USE_PERSISTENT = false; // Insert UCExpr by

    // default

    Block block; // Block that is represented by this Tree

    Subroutine sub; // The Subroutine that we're currently in

    Block next; // The Block after this one

    OperandStack stack; // The program stack for this block

    StmtList stmts; // The statements in the basic block

    Stack savedStack;

    static int stackpos = 0;

    boolean saveValue; // Do we push a StoreExpr on operand stack?

    // Some dup instruction combinations cause saveStack to generate
    // temporaries that clobber other temporaries. So, we use nextindex
    // to ensure that a uniquely-name temporary is created by
    // saveStack. Yes, this will introduce a lot of new temporaries,
    // but expression propagation should eliminate a lot of them.
    private int nextIndex = 0;

    private void db(String s) {
        if (DEBUG)
            System.out.println(s);
    }

    /**
     * Constructor.
     * 
     * @param block
     *            The basic Block of code represented in this Tree.
     * @param predStack
     *            The contents of the operand stack from the previous basic
     *            Block.
     */
    public Tree(Block block, OperandStack predStack) {
        this.block = block;

        if (Tree.DEBUG) {
            System.out.println("    new tree for " + block);
        }

        stack = new OperandStack();

        stmts = new StmtList();

        // The first statement in the Tree is the label indicating the start
        // of the basic Block.
        appendStmt(new LabelStmt(block.label()));

        // Make a copy of predStack
        for (int i = 0; i < predStack.size(); i++) {
            Expr expr = (Expr) predStack.get(i);
            Expr copy = (Expr) expr.clone();
            copy.setDef(null);
            stack.push(copy);
        }
    }

    /**
     * Cleans up this node. Does nothing in this case.
     */
    public void cleanupOnly() {
    }

    /**
     * Add a Collection of local variables to the block. Add an InitStmt to the
     * statement list.
     * 
     * @see LocalExpr
     * @see InitStmt
     */
    public void initLocals(Collection locals) {
        LocalExpr[] t = new LocalExpr[locals.size()];

        if (t.length == 0) {
            return;
        }

        Iterator iter = locals.iterator();

        for (int i = 0; iter.hasNext(); i++) {
            t[i] = (LocalExpr) iter.next();
        }

        addStmt(new InitStmt(t));
    }

    /**
     * Removes a statement from the statement list.
     * 
     * @param stmt
     *            The statement to remove
     */
    public void removeStmt(Stmt stmt) {
        stmts.remove(stmt);
    }

    /**
     * Removes the last non-Label statement from the statement list.
     */
    public void removeLastStmt() {
        ListIterator iter = stmts.listIterator(stmts.size());

        while (iter.hasPrevious()) {
            Stmt s = (Stmt) iter.previous();

            if (s instanceof LabelStmt) {
                continue;
            }

            iter.remove();
            return;
        }
    }

    /**
     * @return The statement list
     */
    public List stmts() {
        return stmts;
    }

    /**
     * StmtList is a linked list of statements. A number of methods are
     * overridden because some adjustments may need to be made to the Nodes in
     * the tree when certain operations are performed.
     */
    class StmtList extends LinkedList {
        static final long serialVersionUID = 1L;

        /**
         * Clear the contents of this statement list.
         */
        public void clear() {
            Iterator iter = iterator();

            while (iter.hasNext()) {
                ((Stmt) iter.next()).cleanup();
            }

            super.clear();
        }

        /**
         * Remove a statement from the list and clean up afterwards.
         */
        public boolean remove(Object o) {
            if (super.remove(o)) {
                ((Stmt) o).cleanup();
                return true;
            }

            return false;
        }

        /**
         * Remove all of the statements in a Collection from this statement
         * list.
         * 
         * @param c
         *            A Collection containing the statements to remove.
         * @return True, if the contents of this statement list changed.
         */
        public boolean removeAll(Collection c) {
            boolean changed = false;

            if (c == this) {
                changed = size() > 0;
                clear();
            } else {
                Iterator iter = c.iterator();

                while (iter.hasNext()) {
                    changed = remove(iter.next()) || changed;
                }
            }

            return changed;
        }

        /**
         * Remove all statements in this list except those that are in a
         * specified Collection.
         * 
         * @param c
         *            The statements to keep.
         * @return True, if the contents of this statement list changed.
         */
        public boolean retainAll(Collection c) {
            boolean changed = false;

            if (c == this) {
                return false;
            }

            Iterator iter = iterator();

            while (iter.hasNext()) {
                if (!c.contains(iter.next())) {
                    changed = true;
                    iter.remove();
                }
            }

            return changed;
        }

        /**
         * Set the value of a given statement.
         * 
         * @param index
         *            Index of the statement to change.
         * @param element
         *            New value of statement at index.
         * 
         * @return Statement previously at position index.
         */
        public Object set(int index, Object element) {
            if (index < size()) {
                Stmt s = (Stmt) get(index);

                if (s != element) {
                    s.cleanup();
                }
            }

            return super.set(index, element);
        }

        /**
         * Removes the statement at index
         * 
         * @return Statement previously at position index.
         */
        public Object remove(int index) {
            Object o = super.remove(index);

            if (o != null) {
                ((Stmt) o).cleanup();
            }

            return o;
        }

        /**
         * Removes statements in a given index range.
         */
        /*
         * public void removeRange(int fromIndex, int toIndex) { int remaining =
         * toIndex - fromIndex;
         * 
         * ListIterator iter = listIterator(fromIndex);
         * 
         * while (iter.hasNext() && remaining-- > 0) { ((Stmt)
         * iter.next()).cleanup(); }
         * 
         * super.removeRange(fromIndex, toIndex); }
         */

        /**
         * @return A ListIterator starting with the first statement in the
         *         statement list.
         */
        public ListIterator listIterator() {
            return listIterator(0);
        }

        /**
         * @return A ListIterator starting a given index.
         */
        public ListIterator listIterator(int index) {
            final ListIterator iter = super.listIterator(index);

            return new ListIterator() {
                Object last = null;

                public boolean hasNext() {
                    return iter.hasNext();
                }

                public Object next() {
                    last = iter.next();
                    return last;
                }

                public boolean hasPrevious() {
                    return iter.hasPrevious();
                }

                public Object previous() {
                    last = iter.previous();
                    return last;
                }

                public int nextIndex() {
                    return iter.nextIndex();
                }

                public int previousIndex() {
                    return iter.previousIndex();
                }

                public void add(Object obj) {
                    Assert.isTrue(obj instanceof Stmt);
                    ((Stmt) obj).setParent(Tree.this);
                    last = null;
                    iter.add(obj);
                }

                public void set(Object obj) {
                    if (last == null) {
                        throw new NoSuchElementException();
                    }

                    Assert.isTrue(obj instanceof Stmt);
                    ((Stmt) obj).setParent(Tree.this);

                    ((Stmt) last).cleanup();
                    last = null;

                    iter.set(obj);
                }

                public void remove() {
                    if (last == null) {
                        throw new NoSuchElementException();
                    }

                    ((Stmt) last).cleanup();
                    last = null;

                    iter.remove();
                }
            };
        }

        /**
         * @return An Iterator over this statement list.
         */
        public Iterator iterator() {
            final Iterator iter = super.iterator();

            return new Iterator() {
                Object last = null;

                public boolean hasNext() {
                    return iter.hasNext();
                }

                public Object next() {
                    last = iter.next();
                    return last;
                }

                public void remove() {
                    if (last == null) {
                        throw new NoSuchElementException();
                    }

                    ((Stmt) last).cleanup();
                    last = null;

                    iter.remove();
                }
            };
        }
    }

    /**
     * Returns the last non-Label statement in the statement list.
     */
    public Stmt lastStmt() {
        ListIterator iter = stmts.listIterator(stmts.size());

        while (iter.hasPrevious()) {
            Stmt s = (Stmt) iter.previous();
            if (Tree.DEBUG)
                System.out.println("Stmt: " + s);

            if (s instanceof LabelStmt) {
                continue;
            }

            return s;
        }
        return null;
    }

    /**
     * Returns the operand stack.
     */
    public OperandStack stack() {
        return stack;
    }

    /**
     * Inserts a statement into the statement list after another given
     * statement.
     * 
     * @param stmt
     *            The statement to add.
     * @param after
     *            The statement after which to add stmt.
     */
    public void addStmtAfter(Stmt stmt, Stmt after) {
        if (Tree.DEBUG) {
            System.out.println("insert: " + stmt + " after " + after);
        }

        ListIterator iter = stmts.listIterator();

        while (iter.hasNext()) {
            Stmt s = (Stmt) iter.next();

            if (s == after) {
                iter.add(stmt);
                stmt.setParent(this);
                return;
            }
        }

        throw new RuntimeException(after + " not found");
    }

    /**
     * Inserts a statement into the statement list before a specified statement.
     * 
     * @param stmt
     *            The statement to insert
     * @param before
     *            The statement before which to add stmt.
     */
    public void addStmtBefore(Stmt stmt, Stmt before) {
        if (Tree.DEBUG) {
            System.out.println("insert: " + stmt + " before " + before);
        }

        ListIterator iter = stmts.listIterator();

        while (iter.hasNext()) {
            Stmt s = (Stmt) iter.next();

            if (s == before) {
                iter.previous();
                iter.add(stmt);
                stmt.setParent(this);
                return;
            }
        }

        throw new RuntimeException(before + " not found");
    }

    /**
     * Add an statement to the statement list before the first non-Label
     * statement.
     * 
     * @param stmt
     *            The statement to add.
     */
    public void prependStmt(Stmt stmt) {
        if (Tree.DEBUG) {
            System.out.println("prepend: " + stmt + " in " + block);
        }

        ListIterator iter = stmts.listIterator();

        while (iter.hasNext()) {
            Stmt s = (Stmt) iter.next();

            if (!(s instanceof LabelStmt)) {
                iter.previous();
                iter.add(stmt);
                stmt.setParent(this);
                return;
            }
        }

        appendStmt(stmt);
    }

    /**
     * When we build the expression tree, there may be items left over on the
     * operand stack. We want to save these items. If USE_STACK is true, then we
     * place these items into stack variables. If USE_STACK is false, then we
     * place the items into local variables.
     */
    private void saveStack() {
        int height = 0;

        for (int i = 0; i < stack.size(); i++) {
            Expr expr = (Expr) stack.get(i);

            if (USE_STACK) {
                // Save to a stack variable only if we'll create a new
                // variable there.
                if (!(expr instanceof StackExpr)
                        || ((StackExpr) expr).index() != height) {

                    StackExpr target = new StackExpr(height, expr.type());

                    // Make a new statement to store the expression that was
                    // part of the stack into memory.
                    Stmt store = new ExprStmt(new StoreExpr(target, expr, expr
                            .type()));

                    appendStmt(store);

                    StackExpr copy = (StackExpr) target.clone();
                    copy.setDef(null);
                    stack.set(i, copy);
                }

            } else {
                if (!(expr instanceof LocalExpr)
                        || !((LocalExpr) expr).fromStack()
                        || ((LocalExpr) expr).index() != height) {

                    LocalExpr target = newStackLocal(nextIndex++, expr.type());

                    Stmt store = new ExprStmt(new StoreExpr(target, expr, expr
                            .type()));

                    appendStmt(store);

                    LocalExpr copy = (LocalExpr) target.clone();
                    copy.setDef(null);
                    stack.set(i, copy);
                }
            }
            height += expr.type().getSize();
        }
    }

    /**
     * Add a statement to this Tree statement list and specify that this is the
     * statement's parent.
     */
    private void appendStmt(Stmt stmt) {
        if (Tree.DEBUG) {
            System.out.println("      append: " + stmt);
        }

        stmt.setParent(this);
        stmts.add(stmt);
    }

    /**
     * Save the contents of the stack and add stmt to the statement list.
     * 
     * @param stmt
     *            A statement to add to the statement list.
     */
    public void addStmt(Stmt stmt) {
        saveStack();
        appendStmt(stmt);
    }

    /**
     * Adds a statement to the statement list before the last jump statement. It
     * is assumed that the last statement in the statement list is a jump
     * statement.
     * 
     * @see JumpStmt
     */
    public void addStmtBeforeJump(Stmt stmt) {
        Stmt last = lastStmt();
        Assert.isTrue(last instanceof JumpStmt, "Last statement of " + block
                + " is " + last + ", not a jump");
        addStmtBefore(stmt, last);
    }

    /**
     * Throw a new ClassFormatException with information about method and class
     * that this basic block is in.
     * 
     * @param msg
     *            String description of the exception.
     * 
     * @see ClassFormatException
     */
    private void throwClassFormatException(String msg) {
        MethodGen method = block.graph().method();

        throw new ClassFormatException("Method " + method.getClassName() + "."
                + method.getName() + " " + method.getReturnType() + ": " + msg);
    }

    /**
     * The last instruction we saw. addInst(Instruction) needs this information.
     */
    Instruction last = null;

    /**
     * Adds an instruction that jumps to another basic block.
     * 
     * @param inst
     *            The instruction to add.
     * @param next
     *            The basic block after the jump. Remember that a jump ends a
     *            basic block.
     * 
     * @see Instruction#isJsr
     * @see Instruction#isConditionalJump
     */
    // Maybe modify to return the InstructionHandle of the added
    // Instruction?
    public void addInstruction(Instruction inst, Block next) {
        Assert.isTrue(inst instanceof JsrInstruction
                || inst instanceof IfInstruction,
                "Wrong addInstruction called with " + inst);
        Assert.isTrue(next != null, "Null next block for " + inst);

        this.next = next;
        addInst(inst);
    }

    /**
     * Adds an instruction that does not change the control flow (a normal
     * instruction).
     * 
     * @param inst
     *            Instruction to add.
     */
    // Maybe modify to return the InstructionHandle of the added
    // Instruction?
    public void addInstruction(Instruction inst) {
        Assert.isTrue(!(inst instanceof JsrInstruction)
                && !(inst instanceof IfInstruction),
                "Wrong addInstruction called with " + inst);

        this.next = null;
        addInst(inst);
    }

    /**
     * Add an instruction such as <i>ret</i> or <i>astore</i> that may involve
     * a subroutine.
     * 
     * @param inst
     *            Instruction to add.
     * @param sub
     *            Subroutine in which inst resides. The <i>ret</i> instruction
     *            always resides in a Subroutine. An <i>astore</i> may store
     *            the return address of a subroutine in a local variable.
     * 
     * @see Instruction#isRet
     */
    // Maybe modify to return the InstructionHandle of the added
    // Instruction?
    public void addInstruction(Instruction inst, Subroutine sub) {
        Assert.isTrue((inst instanceof RET || inst instanceof ASTORE),
                "Wrong addInstruction called with " + inst);

        this.sub = sub;
        this.next = null;
        addInst(inst);
    }

    /**
     * Add a label to the statement list. A label is inserted before a dup
     * statement, but after any other statement.
     * 
     * @param label
     *            Label to add.
     */
    // HUH? WHY ARE WE ADDING A LABEL
    public void addLabel(InstructionHandle label) {
        // Add before a dup, but after any other instruction.
        if (last != null) {
            if ((last instanceof DUP) || (last instanceof DUP2)
                    || (last instanceof DUP_X1) || (last instanceof DUP2_X1)
                    || (last instanceof DUP_X2) || (last instanceof DUP2_X2)) {
            } else {
                addInst(last, false);
                last = null;
            }
        }

        addStmt(new LabelStmt(label));
    }

    /**
     * Private method that adds an instruction to the Tree. The visit method of
     * the instruction is called with this tree as the visitor. The instruction,
     * in turn, calls the appropriate method of this for adding the instruction
     * to the statement list.
     * 
     * @param inst
     *            The <tt>Instruction</tt> to add to the <tt>Tree</tt>
     * @param saveValue
     *            Do we save expressions on the operand stack to stack variables
     *            or local variables.
     */
    // Maybe modify to return the InstructionHandle of the added
    // Instruction?
    private void addInst(Instruction inst, boolean saveValue) {

        if (Tree.DEBUG) {
            // Print the contents of the stack
            int i = 0;
            while (i < stack.size()) {
                Expr exp = stack.peek(i);
                i = i + exp.type().getSize();
                System.out.println((i > 0 ? "-" + i : " " + i) + ": " + exp);
            }
        }

        if (Tree.DEBUG) {
            System.out.println("    add " + inst + " save=" + saveValue);
        }
        try {
            this.saveValue = saveValue;

            if (FLATTEN) {
                saveStack();
            }
            visitInst(inst);

        } catch (EmptyStackException e) {
            throwClassFormatException("Empty operand stack at " + inst);
            return;
        }
    }

    /**
     * Private method that adds an instruction to the tree. Before dispatching
     * the addition off to addInst(Instruction, boolean), it optimizes(?) dup
     * instructions as outlined below:
     */
    // (S0, S1) := dup(X)
    // L := S1 --> use (L := X)
    // use S0
    //
    // (S0, S1, S2) := dup_x1(X, Y)
    // S1.f := S2 --> use (X.f := Y)
    // use S0
    //
    // (S0, S1, S2, S3) := dup_x2(X, Y, Z)
    // S1[S2] := S3 --> use (X[Y] := Z)
    // use S0
    //
    private void addInst(Instruction inst) {
        // Build the tree, trying to convert dup-stores
        if (last == null) {
            last = inst;

        } else {
            if (last instanceof DUP) {
                if ((inst instanceof ASTORE) || (inst instanceof FSTORE)
                        || (inst instanceof ISTORE)
                        || (inst instanceof PUTSTATIC)) {
                    addInst(inst, true);
                    last = null;
                }
            } else if (last instanceof DUP2) {
                if ((inst instanceof DSTORE) || (inst instanceof LSTORE)
                        || (inst instanceof PUTSTATIC)) {
                    addInst(inst, true);
                    last = null;
                }
            } else if (last instanceof DUP_X1) {
                if (inst instanceof PUTFIELD) {// PROBLEM HEREcase
                    // opcx_putfield_nowb:
                    addInst(inst, true);
                    last = null;
                }
            } else if (last instanceof DUP2_X1) {
                if (inst instanceof PUTFIELD) {// PROBLEM HEREcase
                    // opcx_putfield_nowb:
                    addInst(inst, true);
                    last = null;
                }
            } else if (last instanceof DUP_X2) {
                if ((inst instanceof AASTORE) || (inst instanceof BASTORE)
                        || (inst instanceof CASTORE)
                        || (inst instanceof FASTORE)
                        || (inst instanceof IASTORE)
                        || (inst instanceof SASTORE)) {
                    addInst(inst, true);
                    last = null;
                }
            } else if (last instanceof DUP2_X2) {
                if ((inst instanceof DASTORE) || (inst instanceof LASTORE)) {
                    addInst(inst, true);
                    last = null;
                }
            }

            if (last != null) {
                addInst(last, false);
                last = inst;

            }
        }

        // We should have dealt with the old last instruction
        Assert.isTrue(last == null || last == inst);

        if (inst instanceof BranchInstruction
                || inst instanceof ReturnInstruction || inst instanceof RET
                || inst instanceof JsrInstruction || inst instanceof Select
                || inst instanceof ATHROW) {
            addInst(inst, false);
            last = null;
        }
    }

    /**
     * Returns a new StackExpr for the top of the operand stack.
     */
    public StackExpr newStack(Type type) {
        return new StackExpr(stackpos++, type);
    }

    /**
     * Returns a new LocalExpr that represents an element of the stack. They are
     * created when the USE_STACK flag is not set.
     * 
     * @param index
     *            Stack index of variable.
     * @param type
     *            The type of the LocalExpr
     */
    public LocalExpr newStackLocal(int index, Type type) {
        if (index >= nextIndex) {
            nextIndex = index + 1;
        }
        return new LocalExpr(index, true, type);
    }

    /**
     * Returns a new LocalExpr that is not allocated on the stack.
     * 
     * @param index
     *            Stack index of variable.
     * @param type
     *            The type of the LocalExpr
     */
    public LocalExpr newLocal(int index, Type type) {
        return new LocalExpr(index, false, type);
    }

    /**
     * Returns a local variable (<tt>LocalExpr</tt>) located in this method.
     * 
     * @param type
     *            The type of the new LocalExpr.
     */
    public LocalExpr newLocal(Type type) {
        LocalVariableGen var = block.graph().method().addLocalVariable("",
                type, null, null);
        var.setName("newLocalVar" + var.getIndex());
        // presuming it is bad not to have a name....
        return new LocalExpr(var.getIndex(), type);
    }

    /**
     * Returns a String representation of this Tree.
     */
    public String toString() {
        String x = "(TREE " + block + " stack=";

        for (int i = 0; i < stack.size(); i++) {
            Expr expr = (Expr) stack.get(i);
            x += expr.type();
        }

        return x + ")";
    }

    public void visitInst(Instruction inst) {
        if (Tree.DEBUG)
            System.out.println("Opcode = " + inst.getOpcode());

        inst.accept(this);
    }

    /**
     * Adds no statements to the statement list.
     */
    public void visitNOP(NOP inst) {
    }

    public void visitDCONST(DCONST inst) {
        Expr top = new ConstantExpr((Double) inst.getValue(), Type.DOUBLE);
        stack.push(top);
    }

    public void visitFCONST(FCONST inst) {
        Expr top = new ConstantExpr((Float) inst.getValue(), Type.FLOAT);
        stack.push(top);
    }

    public void visitICONST(ICONST inst) {
        Expr top = new ConstantExpr((Integer) inst.getValue(), Type.INT);
        stack.push(top);
    }

    public void visitACONST_NULL(ACONST_NULL inst) {
        Expr top = new ConstantExpr(null, Type.OBJECT);
        stack.push(top);
    }

    public void visitLCONST(LCONST inst) {
        Expr top = new ConstantExpr((Long) inst.getValue(), Type.LONG);
        stack.push(top);
    }

    public void visitBIPUSH(BIPUSH inst) {
        Expr top = new ConstantExpr(inst.getValue(), Type.BYTE);
        stack.push(top);
    }

    public void visitSIPUSH(SIPUSH inst) {
        Expr top = new ConstantExpr(inst.getValue(), Type.SHORT);
        stack.push(top);
    }

    /**
     * Pushes a ConstantExpr onto the operand stack.
     * 
     * @see ConstantExpr
     */
    public void visitLDC2_W(LDC2_W inst) {
        ConstantPoolGen cp = block.graph().method().getConstantPool();
        Object value = inst.getValue(cp);
        Type type;

        if (value instanceof Long) {
            type = Type.LONG;
        } else if (value instanceof Double) {
            type = Type.DOUBLE;
        } else {
            throw new ClassFormatException("Illegal constant type: "
                    + value.getClass().getName() + ": " + value);
            // return;
        }

        Expr top = new ConstantExpr(value, type);
        stack.push(top);
    }

    /**
     * Pushes a ConstantExpr onto the operand stack.
     * 
     * @see ConstantExpr
     */
    public void visitLDC(LDC inst) {
        ConstantPoolGen cp = block.graph().method().getConstantPool();
        Object value = inst.getValue(cp);
        Type type;

        if (value == null) {
            type = Type.OBJECT;
        } else if (value instanceof Integer) {
            type = Type.INT;
        }
        // else if (value instanceof Long) {
        // type = Type.LONG;
        // }
        else if (value instanceof Float) {
            type = Type.FLOAT;
        }
        // else if (value instanceof Double) {
        // type = Type.DOUBLE;
        // }
        else if (value instanceof String) {
            type = Type.STRING;
        } else {
            throw new ClassFormatException("Illegal constant type: "
                    + value.getClass().getName() + ": " + value);
            // return;
        }

        Expr top = new ConstantExpr(value, type);
        stack.push(top);
    }

    /**
     * All <tt>visit_<i>x</i>load</tt> push a LocalExpr onto the operand
     * stack.
     * 
     * @see LocalExpr
     */

    public void visitILOAD(ILOAD iInst) {
        Expr top = new LocalExpr(iInst.getIndex(), Type.INT);
        stack.push(top);
        db("      iload: " + top);
    }

    public void visitLLOAD(LLOAD iInst) {
        Expr top = new LocalExpr(iInst.getIndex(), Type.LONG);
        stack.push(top);
        db("      lload: " + top);
    }

    public void visitFLOAD(FLOAD iInst) {
        Expr top = new LocalExpr(iInst.getIndex(), Type.FLOAT);
        stack.push(top);
        db("      fload: " + top);
    }

    public void visitDLOAD(DLOAD iInst) {
        Expr top = new LocalExpr(iInst.getIndex(), Type.DOUBLE);
        stack.push(top);
        db("      dload: " + top);
    }

    public void visitALOAD(ALOAD iInst) {
        Expr top = new LocalExpr(iInst.getIndex(), Type.OBJECT);
        stack.push(top);
        db("      aload: " + top);
    }

    /**
     * All <tt>visit_<i>x</i>aload</tt> push an ArrayRefExpr onto the
     * operand stack.
     */
    public void visitIALOAD(IALOAD inst) {
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.INT, 1));
        // 0 should do for array dimensions I don't think it is important
        Expr top = new ArrayRefExpr(array, index, Type.INT, Type.INT);
        stack.push(top);
    }

    public void visitLALOAD(LALOAD inst) {
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.LONG, 1));
        Expr top = new ArrayRefExpr(array, index, Type.LONG, Type.LONG);
        stack.push(top);
    }

    public void visitFALOAD(FALOAD inst) {
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.FLOAT, 1));
        Expr top = new ArrayRefExpr(array, index, Type.FLOAT, Type.FLOAT);
        stack.push(top);
    }

    public void visitDALOAD(DALOAD inst) {
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.DOUBLE, 1));
        Expr top = new ArrayRefExpr(array, index, Type.DOUBLE, Type.DOUBLE);
        stack.push(top);
    }

    public void visitAALOAD(AALOAD inst) {
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.OBJECT, 1));
        Expr top = new ArrayRefExpr(array, index, Type.OBJECT, Type.OBJECT);
        stack.push(top);
    }

    public void visitBALOAD(BALOAD inst) {
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.BYTE, 1));
        Expr top = new ArrayRefExpr(array, index, Type.BYTE, Type.BYTE);
        stack.push(top);
    }

    public void visitCALOAD(CALOAD inst) {
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.CHAR, 1));
        Expr top = new ArrayRefExpr(array, index, Type.CHAR, Type.CHAR);
        stack.push(top);
    }

    public void visitSALOAD(SALOAD inst) {
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.SHORT, 1));
        Expr top = new ArrayRefExpr(array, index, Type.SHORT, Type.SHORT);
        stack.push(top);
    }

    /**
     * Deals with an expression that stores a value. It either pushes it on the
     * operand stack or adds a statement depending on the value of saveValue.
     * 
     * @param target
     *            The location to where we are storing the value.
     * @param expr
     *            The expression whose value we are storing.
     */
    private void addStore(MemExpr target, Expr expr) {
        // Stmt last = lastStmt();

        if (saveValue) {
            stack.push(new StoreExpr(target, expr, expr.type()));

        } else {
            addStmt(new ExprStmt(new StoreExpr(target, expr, expr.type())));
        }
    }

    /**
     * All <tt>visit_<i>x</i>store</tt> add a LocalExpr statement to the
     * statement list.
     */
    public void visitISTORE(ISTORE iInst) {
        Expr expr = (Expr) stack.pop(Type.INT);
        LocalExpr target = new LocalExpr(iInst.getIndex(), expr.type());
        addStore(target, expr);
    }

    public void visitLSTORE(LSTORE iInst) {
        Expr expr = (Expr) stack.pop(Type.LONG);
        LocalExpr target = new LocalExpr(iInst.getIndex(), expr.type());
        addStore(target, expr);
    }

    public void visitFSTORE(FSTORE iInst) {
        Expr expr = (Expr) stack.pop(Type.FLOAT);
        LocalExpr target = new LocalExpr(iInst.getIndex(), expr.type());
        addStore(target, expr);
    }

    public void visitDSTORE(DSTORE iInst) {
        Expr expr = (Expr) stack.pop(Type.DOUBLE);
        LocalExpr target = new LocalExpr(iInst.getIndex(), expr.type());
        addStore(target, expr);
    }

    /**
     * Visit an <i>astore</i> instruction. If the type of the operand to the
     * instruction is an address add an AddressStoreStmt to the tree, else add a
     * StoreStmt to the tree consisting of a LocalExpr and the top Expr on the
     * operand stack.
     * 
     * @see AddressStoreStmt
     * @see LocalExpr
     * @see StoreExpr
     */
    public void visitASTORE(ASTORE iInst) {

        Expr expr = (Expr) stack.peek();

        if (expr.type() instanceof ReturnaddressType) {
            ReturnaddressType rat = (ReturnaddressType) expr.type();
            Assert.isTrue(sub != null);
            Assert.isTrue(!saveValue);
            expr = (Expr) stack.pop(ReturnaddressType.NO_TARGET);
            // something refering to returnaddressTypes???Can't think of
            // anything better
            sub.setReturnAddress(rat);
            addStmt(new AddressStoreStmt(sub));
        } else {
            expr = (Expr) stack.pop(Type.OBJECT);
            LocalExpr target = new LocalExpr(iInst.getIndex(), expr.type());
            addStore(target, expr);
        }
    }

    public void visitIASTORE(IASTORE inst) {
        Expr value = (Expr) stack.pop(Type.INT);
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.INT, 1));
        ArrayRefExpr target = new ArrayRefExpr(array, index, Type.INT, Type.INT);
        addStore(target, value);
    }

    public void visitLASTORE(LASTORE inst) {
        Expr value = (Expr) stack.pop(Type.LONG);
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.LONG, 1));
        ArrayRefExpr target = new ArrayRefExpr(array, index, Type.LONG,
                Type.LONG);
        addStore(target, value);
    }

    public void visitFASTORE(FASTORE inst) {
        Expr value = (Expr) stack.pop(Type.FLOAT);
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.FLOAT, 1));
        ArrayRefExpr target = new ArrayRefExpr(array, index, Type.FLOAT,
                Type.FLOAT);
        addStore(target, value);
    }

    public void visitDASTORE(DASTORE inst) {
        Expr value = (Expr) stack.pop(Type.DOUBLE);
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.DOUBLE, 1));
        ArrayRefExpr target = new ArrayRefExpr(array, index, Type.DOUBLE,
                Type.DOUBLE);
        addStore(target, value);
    }

    public void visitAASTORE(AASTORE inst) {
        Expr value = (Expr) stack.pop(Type.OBJECT);
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.OBJECT, 1));
        ArrayRefExpr target = new ArrayRefExpr(array, index, Type.OBJECT,
                Type.OBJECT);
        addStore(target, value);
    }

    public void visitBASTORE(BASTORE inst) {
        Expr value = (Expr) stack.pop(Type.BYTE);
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.BYTE, 1));
        ArrayRefExpr target = new ArrayRefExpr(array, index, Type.BYTE,
                Type.BYTE);
        addStore(target, value);
    }

    public void visitCASTORE(CASTORE inst) {
        Expr value = (Expr) stack.pop(Type.CHAR);
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.CHAR, 1));
        ArrayRefExpr target = new ArrayRefExpr(array, index, Type.CHAR,
                Type.CHAR);
        addStore(target, value);
    }

    public void visitSASTORE(SASTORE inst) {
        Expr value = (Expr) stack.pop(Type.SHORT);
        Expr index = (Expr) stack.pop(Type.INT);
        Expr array = (Expr) stack.pop(new ArrayType(Type.SHORT, 1));
        ArrayRefExpr target = new ArrayRefExpr(array, index, Type.SHORT,
                Type.SHORT);
        addStore(target, value);
    }

    /**
     * Pop the expression off the top of the stack and add it as an ExprStmt to
     * the statement list.
     * 
     * @see ExprStmt
     */
    public void visitPOP(POP inst) {
        Expr expr = stack.pop1();
        addStmt(new ExprStmt(expr));
    }

    public void visitPOP2(POP2 inst) {
        Expr[] expr = stack.pop2();

        if (expr.length == 1) {
            addStmt(new ExprStmt(expr[0]));
        } else {
            addStmt(new ExprStmt(expr[0]));
            addStmt(new ExprStmt(expr[1]));
        }
    }

    /**
     * When processing the <i>dup</i> instructions one of two situations can
     * occur. If the <tt>USE_STACK</tt> flag is set, then a StackManipStmt is
     * created to represent the transformation that the <i>dup</i> instruction
     * performs on the stack. If the <tt>USE_STACK</tt> flag is not set, then
     * the transformation is simulated by creating new local variables
     * containing the appropriate element of the stack.
     * 
     * @see LocalExpr
     * @see StackExpr
     * @see StackManipStmt
     */
    public void visitDUP(DUP inst) {
        // 0 -> 0 0

        db("      dup");

        if (USE_STACK) {
            saveStack();

            StackExpr s0 = (StackExpr) stack.pop1();

            StackExpr[] s = new StackExpr[] { s0 };
            manip(s, new int[] { 0, 0 }, StackManipStmt.DUP);

        } else {
            Expr s0 = stack.pop1();

            LocalExpr t0 = newStackLocal(stack.height(), s0.type());

            db("        s0: " + s0);
            db("        t0: " + t0);

            if (!t0.equalsExpr(s0)) {
                db("          t0 <- s0");
                addStore(t0, s0);
            }

            Expr copy = (Expr) t0.clone();
            copy.setDef(null);
            stack.push(copy);

            copy = (Expr) t0.clone();
            copy.setDef(null);
            stack.push(copy);
        }
    }

    public void visitDUP_X1(DUP_X1 inst) {
        // 0 1 -> 1 0 1

        if (USE_STACK) {
            saveStack();

            StackExpr s1 = (StackExpr) stack.pop1();
            StackExpr s0 = (StackExpr) stack.pop1();

            StackExpr[] s = new StackExpr[] { s0, s1 };
            manip(s, new int[] { 1, 0, 1 }, StackManipStmt.DUP_X1);

        } else {
            Expr s1 = stack.pop1();
            Expr s0 = stack.pop1();

            LocalExpr t0 = newStackLocal(stack.height(), s0.type());
            LocalExpr t1 = newStackLocal(stack.height() + 1, s1.type());

            if (!t0.equalsExpr(s0)) {
                addStore(t0, s0);
            }

            if (!t1.equalsExpr(s1)) {
                addStore(t1, s1);
            }

            Expr copy = (Expr) t1.clone();
            copy.setDef(null);
            stack.push(copy);

            copy = (Expr) t0.clone();
            copy.setDef(null);
            stack.push(copy);

            copy = (Expr) t1.clone();
            copy.setDef(null);
            stack.push(copy);
        }
    }

    public void visitDUP_X2(DUP_X2 inst) {
        // 0 1 2 -> 2 0 1 2
        // 0-1 2 -> 2 0-1 2

        db("      dup_x2");

        if (USE_STACK) {
            saveStack();

            StackExpr s2 = (StackExpr) stack.pop1();
            Expr[] s01 = stack.pop2();

            if (s01.length == 2) {
                // 0 1 2 -> 2 0 1 2
                StackExpr[] s = new StackExpr[] { (StackExpr) s01[0],
                        (StackExpr) s01[1], (StackExpr) s2 };
                manip(s, new int[] { 2, 0, 1, 2 }, StackManipStmt.DUP_X2);
            } else {
                // 0-1 2 -> 2 0-1 2
                StackExpr[] s = new StackExpr[] { (StackExpr) s01[0],
                        (StackExpr) s2 };
                manip(s, new int[] { 1, 0, 1 }, StackManipStmt.DUP_X2);
            }

        } else {
            Expr s2 = stack.pop1();
            Expr[] s01 = stack.pop2();

            db("        s2: " + s2);
            db("        s01: " + s01[0] + (s01.length > 1 ? " " + s01[1] : ""));

            if (s01.length == 2) {
                // 0 1 2 -> 2 0 1 2
                LocalExpr t0 = newStackLocal(stack.height(), s01[0].type());
                LocalExpr t1 = newStackLocal(stack.height() + 1, s01[1].type());
                LocalExpr t2 = newStackLocal(stack.height() + 2, s2.type());

                db("        t0: " + t0);
                db("        t1: " + t1);
                db("        t2: " + t2);

                if (!t0.equalsExpr(s01[0])) {
                    db("          t0 <- s01[0]");
                    addStore(t0, s01[0]);
                }

                if (!t1.equalsExpr(s01[1])) {
                    db("          t1 <- s01[1]");
                    addStore(t1, s01[1]);
                }

                if (!t2.equalsExpr(s2)) {
                    db("          t2 <- s2");
                    addStore(t2, s2);
                }

                Expr copy = (Expr) t2.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t0.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t1.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t2.clone();
                copy.setDef(null);
                stack.push(copy);

            } else {
                // 0-1 2 -> 2 0-1 2
                LocalExpr t0 = newStackLocal(stack.height(), s01[0].type());
                LocalExpr t2 = newStackLocal(stack.height() + 2, s2.type());

                if (!t0.equalsExpr(s01[0])) {
                    addStore(t0, s01[0]);
                }

                if (!t2.equalsExpr(s2)) {
                    addStore(t2, s2);
                }

                Expr copy = (Expr) t2.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t0.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t2.clone();
                copy.setDef(null);
                stack.push(copy);
            }
        }
    }

    public void visitDUP2(DUP2 inst) {
        // 0 1 -> 0 1 0 1
        // 0-1 -> 0-1 0-1

        if (USE_STACK) {
            saveStack();

            Expr[] s01 = stack.pop2();

            if (s01.length == 1) {
                // 0-1 -> 0-1 0-1
                StackExpr[] s = new StackExpr[] { (StackExpr) s01[0] };
                manip(s, new int[] { 0, 0 }, StackManipStmt.DUP2);
            } else {
                // 0 1 -> 0 1 0 1

                Assert.isTrue(s01.length == 2);

                StackExpr[] s = new StackExpr[] { (StackExpr) s01[0],
                        (StackExpr) s01[1] };
                manip(s, new int[] { 0, 1, 0, 1 }, StackManipStmt.DUP2);
            }
        } else {
            Expr[] s01 = stack.pop2();

            if (s01.length == 1) {
                // 0-1 -> 0-1 0-1
                LocalExpr t0 = newStackLocal(stack.height(), s01[0].type());

                if (!t0.equalsExpr(s01[0])) {
                    addStore(t0, s01[0]);
                }

                Expr copy = (Expr) t0.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t0.clone();
                copy.setDef(null);
                stack.push(copy);
            } else {
                // 0 1 -> 0 1 0 1
                LocalExpr t0 = newStackLocal(stack.height(), s01[0].type());
                LocalExpr t1 = newStackLocal(stack.height() + 1, s01[1].type());

                if (!t0.equalsExpr(s01[0])) {
                    addStore(t0, s01[0]);
                }

                if (!t1.equalsExpr(s01[1])) {
                    addStore(t1, s01[1]);
                }

                Expr copy = (Expr) t0.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t1.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t0.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t1.clone();
                copy.setDef(null);
                stack.push(copy);
            }
        }
    }

    public void visitDUP2_X1(DUP2_X1 inst) {
        // 0 1 2 -> 1 2 0 1 2
        // 0 1-2 -> 1-2 0 1-2

        if (USE_STACK) {
            saveStack();

            Expr[] s12 = stack.pop2();
            StackExpr s0 = (StackExpr) stack.pop1();

            if (s12.length == 2) {
                // 0 1 2 -> 1 2 0 1 2
                StackExpr[] s = new StackExpr[] { (StackExpr) s0,
                        (StackExpr) s12[0], (StackExpr) s12[1] };
                manip(s, new int[] { 1, 2, 0, 1, 2 }, StackManipStmt.DUP2_X1);
            } else {
                // 0 1-2 -> 1-2 0 1-2
                StackExpr[] s = new StackExpr[] { (StackExpr) s0,
                        (StackExpr) s12[0] };
                manip(s, new int[] { 1, 0, 1 }, StackManipStmt.DUP2_X1);
            }
        } else {
            Expr[] s12 = stack.pop2();
            StackExpr s0 = (StackExpr) stack.pop1();

            if (s12.length == 2) {
                // 0 1 2 -> 1 2 0 1 2
                LocalExpr t0 = newStackLocal(stack.height(), s0.type());
                LocalExpr t1 = newStackLocal(stack.height() + 1, s12[0].type());
                LocalExpr t2 = newStackLocal(stack.height() + 2, s12[1].type());

                if (!t0.equalsExpr(s0)) {
                    addStore(t0, s0);
                }

                if (!t1.equalsExpr(s12[0])) {
                    addStore(t1, s12[0]);
                }

                if (!t2.equalsExpr(s12[1])) {
                    addStore(t2, s12[1]);
                }

                Expr copy = (Expr) t1.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t2.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t0.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t1.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t2.clone();
                copy.setDef(null);
                stack.push(copy);
            } else {
                // 0 1-2 -> 1-2 0 1-2
                LocalExpr t0 = newStackLocal(stack.height(), s0.type());
                LocalExpr t1 = newStackLocal(stack.height() + 1, s12[0].type());

                if (!t0.equalsExpr(s0)) {
                    addStore(t0, s0);
                }

                if (!t1.equalsExpr(s12[0])) {
                    addStore(t1, s12[0]);
                }

                Expr copy = (Expr) t1.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t0.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t1.clone();
                copy.setDef(null);
                stack.push(copy);
            }
        }
    }

    public void visitDUP2_X2(DUP2_X2 inst) {
        // 0 1 2 3 -> 2 3 0 1 2 3
        // 0 1 2-3 -> 2-3 0 1 2-3
        // 0-1 2 3 -> 2 3 0-1 2 3
        // 0-1 2-3 -> 2-3 0-1 2-3

        if (USE_STACK) {
            saveStack();

            Expr[] s23 = stack.pop2();
            Expr[] s01 = stack.pop2();

            if (s01.length == 2 && s23.length == 2) {
                // 0 1 2 3 -> 2 3 0 1 2 3
                StackExpr[] s = new StackExpr[] { (StackExpr) s01[0],
                        (StackExpr) s01[1], (StackExpr) s23[0],
                        (StackExpr) s23[1] };
                manip(s, new int[] { 2, 3, 0, 1, 2, 3 }, StackManipStmt.DUP2_X2);
            } else if (s01.length == 2 && s23.length == 1) {
                // 0 1 2-3 -> 2-3 0 1 2-3
                StackExpr[] s = new StackExpr[] { (StackExpr) s01[0],
                        (StackExpr) s01[1], (StackExpr) s23[0] };
                manip(s, new int[] { 2, 0, 1, 2 }, StackManipStmt.DUP2_X2);
            } else if (s01.length == 1 && s23.length == 2) {
                // 0-1 2 3 -> 2 3 0-1 2 3
                StackExpr[] s = new StackExpr[] { (StackExpr) s01[0],
                        (StackExpr) s23[0], (StackExpr) s23[1] };
                manip(s, new int[] { 1, 2, 0, 1, 2 }, StackManipStmt.DUP2_X2);
            } else if (s01.length == 1 && s23.length == 2) {
                // 0-1 2-3 -> 2-3 0-1 2-3
                StackExpr[] s = new StackExpr[] { (StackExpr) s01[0],
                        (StackExpr) s23[0] };
                manip(s, new int[] { 1, 0, 1 }, StackManipStmt.DUP2_X2);
            }
        } else {
            Expr[] s23 = stack.pop2();
            Expr[] s01 = stack.pop2();

            if (s01.length == 2 && s23.length == 2) {
                // 0 1 2 3 -> 2 3 0 1 2 3
                LocalExpr t0 = newStackLocal(stack.height(), s01[0].type());
                LocalExpr t1 = newStackLocal(stack.height() + 1, s01[1].type());
                LocalExpr t2 = newStackLocal(stack.height() + 2, s23[0].type());
                LocalExpr t3 = newStackLocal(stack.height() + 3, s23[1].type());

                if (!t0.equalsExpr(s01[0])) {
                    addStore(t0, s01[0]);
                }

                if (!t1.equalsExpr(s01[1])) {
                    addStore(t1, s01[1]);
                }

                if (!t2.equalsExpr(s23[0])) {
                    addStore(t2, s23[0]);
                }

                if (!t3.equalsExpr(s23[1])) {
                    addStore(t3, s23[1]);
                }

                Expr copy = (Expr) t2.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t3.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t0.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t1.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t2.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t3.clone();
                copy.setDef(null);
                stack.push(copy);
            } else if (s01.length == 2 && s23.length == 1) {
                // 0 1 2-3 -> 2-3 0 1 2-3
                LocalExpr t0 = newStackLocal(stack.height(), s01[0].type());
                LocalExpr t1 = newStackLocal(stack.height() + 1, s01[1].type());
                LocalExpr t2 = newStackLocal(stack.height() + 2, s23[0].type());

                if (!t0.equalsExpr(s01[0])) {
                    addStore(t0, s01[0]);
                }

                if (!t1.equalsExpr(s01[1])) {
                    addStore(t1, s01[1]);
                }

                if (!t2.equalsExpr(s23[0])) {
                    addStore(t2, s23[0]);
                }

                Expr copy = (Expr) t2.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t0.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t1.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t2.clone();
                copy.setDef(null);
                stack.push(copy);
            } else if (s01.length == 1 && s23.length == 2) {
                // 0-1 2 3 -> 2 3 0-1 2 3
                LocalExpr t0 = newStackLocal(stack.height(), s01[0].type());
                LocalExpr t2 = newStackLocal(stack.height() + 2, s23[0].type());
                LocalExpr t3 = newStackLocal(stack.height() + 3, s23[1].type());

                if (!t0.equalsExpr(s01[0])) {
                    addStore(t0, s01[0]);
                }

                if (!t2.equalsExpr(s23[0])) {
                    addStore(t2, s23[0]);
                }

                if (!t3.equalsExpr(s23[1])) {
                    addStore(t3, s23[1]);
                }

                Expr copy = (Expr) t2.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t3.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t0.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t2.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t3.clone();
                copy.setDef(null);
                stack.push(copy);
            } else if (s01.length == 1 && s23.length == 2) {
                // 0-1 2-3 -> 2-3 0-1 2-3
                LocalExpr t0 = newStackLocal(stack.height(), s01[0].type());
                LocalExpr t2 = newStackLocal(stack.height() + 2, s23[0].type());

                if (!t0.equalsExpr(s01[0])) {
                    addStore(t0, s01[0]);
                }

                if (!t2.equalsExpr(s23[0])) {
                    addStore(t2, s23[0]);
                }

                Expr copy = (Expr) t2.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t0.clone();
                copy.setDef(null);
                stack.push(copy);

                copy = (Expr) t2.clone();
                copy.setDef(null);
                stack.push(copy);
            }
        }
    }

    public void visitSWAP(SWAP inst) {
        // 0 1 -> 1 0

        if (USE_STACK) {
            saveStack();

            StackExpr s1 = (StackExpr) stack.pop1();
            StackExpr s0 = (StackExpr) stack.pop1();

            StackExpr[] s = new StackExpr[] { (StackExpr) s0, (StackExpr) s1 };
            manip(s, new int[] { 1, 0 }, StackManipStmt.SWAP);
        } else {
            Expr s1 = stack.pop1();
            Expr s0 = stack.pop1();

            LocalExpr t0 = newStackLocal(stack.height(), s0.type());
            LocalExpr t1 = newStackLocal(stack.height() + 1, s1.type());

            if (!t0.equalsExpr(s0)) {
                addStore(t0, s0);
            }

            if (!t1.equalsExpr(s1)) {
                addStore(t1, s1);
            }

            Expr copy = (Expr) t1.clone();
            copy.setDef(null);
            stack.push(copy);

            copy = (Expr) t0.clone();
            copy.setDef(null);
            stack.push(copy);
        }
    }

    /**
     * Produces a StackManipStmt that represents how the stack is changed as a
     * result of a dup instruction. It should only be used when USE_STACK is
     * true.
     * <p>
     * dup instructions change the top n elements of the JVM stack. This method
     * takes the original top n elements of the stack, an integer array
     * representing the transformation (for instance, if s[0] = 1, then the top
     * element of the new stack should contain the second-from-the-top element
     * of the old stack), and integer representing the dup instruction.
     * 
     * @param source
     *            The interesting part of the stack before the dup instruction
     *            is executed.
     * @param s
     *            An integer array representing the new order of the stack.
     * @param kind
     *            The kind of stack manipulation taking place. (e.g.
     *            StackManipStmt.DUP_X1)
     * 
     * @see StackManipStmt
     */
    private void manip(StackExpr[] source, int[] s, int kind) {
        Assert.isTrue(USE_STACK);

        int height = 0; // Height of the stack

        // Calculate current height of the stack. Recall that the stack
        // elements in source have already been popped off the stack.
        for (int i = 0; i < stack.size(); i++) {
            Expr expr = (Expr) stack.get(i);
            height += expr.type().getSize();
        }

        // Create the new portion of the stack. Make new StackExpr
        // representing the stack after the dup instruction. Push those
        // new StackExprs onto the operand stack. Finally, create a
        // StackManipStmt that represent the transformation of the old
        // stack (before dup instruction) to the new stack (after dup
        // instruction).
        StackExpr[] target = new StackExpr[s.length];

        for (int i = 0; i < s.length; i++) {
            target[i] = new StackExpr(height, source[s[i]].type());
            StackExpr copy = (StackExpr) target[i].clone();
            copy.setDef(null);
            stack.push(copy);
            height += target[i].type().getSize();
        }

        appendStmt(new StackManipStmt(target, source, kind));
    }

    /**
     * All <tt>visit_<i>x</i>add</tt>, <tt>visit_<i>x</i>sub</tt>,
     * <tt>visit_<i>x</i>mul</tt>, <tt>visit_<i>x</i>div</tt>, etc.
     * push an ArithExpr onto the operand stack.
     * 
     * @see ArithExpr
     */
    public void visitIADD(IADD inst) {
        Expr right = (Expr) stack.pop(Type.INT);
        Expr left = (Expr) stack.pop(Type.INT);
        Expr top = new ArithExpr(ArithExpr.ADD, left, right, Type.INT);
        stack.push(top);
    }

    public void visitLADD(LADD inst) {
        Expr right = (Expr) stack.pop(Type.LONG);
        Expr left = (Expr) stack.pop(Type.LONG);
        Expr top = new ArithExpr(ArithExpr.ADD, left, right, Type.LONG);
        stack.push(top);
    }

    public void visitFADD(FADD inst) {
        Expr right = (Expr) stack.pop(Type.FLOAT);
        Expr left = (Expr) stack.pop(Type.FLOAT);
        Expr top = new ArithExpr(ArithExpr.ADD, left, right, Type.FLOAT);
        stack.push(top);
    }

    public void visitDADD(DADD inst) {
        Expr right = (Expr) stack.pop(Type.DOUBLE);
        Expr left = (Expr) stack.pop(Type.DOUBLE);
        Expr top = new ArithExpr(ArithExpr.ADD, left, right, Type.DOUBLE);
        stack.push(top);
    }

    public void visitISUB(ISUB inst) {
        Expr right = (Expr) stack.pop(Type.INT);
        Expr left = (Expr) stack.pop(Type.INT);
        Expr top = new ArithExpr(ArithExpr.SUB, left, right, Type.INT);
        stack.push(top);
    }

    public void visitLSUB(LSUB inst) {
        Expr right = (Expr) stack.pop(Type.LONG);
        Expr left = (Expr) stack.pop(Type.LONG);
        Expr top = new ArithExpr(ArithExpr.SUB, left, right, Type.LONG);
        stack.push(top);
    }

    public void visitFSUB(FSUB inst) {
        Expr right = (Expr) stack.pop(Type.FLOAT);
        Expr left = (Expr) stack.pop(Type.FLOAT);
        Expr top = new ArithExpr(ArithExpr.SUB, left, right, Type.FLOAT);
        stack.push(top);
    }

    public void visitDSUB(DSUB inst) {
        Expr right = (Expr) stack.pop(Type.DOUBLE);
        Expr left = (Expr) stack.pop(Type.DOUBLE);
        Expr top = new ArithExpr(ArithExpr.SUB, left, right, Type.DOUBLE);
        stack.push(top);
    }

    public void visitIMUL(IMUL inst) {
        Expr right = (Expr) stack.pop(Type.INT);
        Expr left = (Expr) stack.pop(Type.INT);
        Expr top = new ArithExpr(ArithExpr.MUL, left, right, Type.INT);
        stack.push(top);
    }

    public void visitLMUL(LMUL inst) {
        Expr right = (Expr) stack.pop(Type.LONG);
        Expr left = (Expr) stack.pop(Type.LONG);
        Expr top = new ArithExpr(ArithExpr.MUL, left, right, Type.LONG);
        stack.push(top);
    }

    public void visitFMUL(FMUL inst) {
        Expr right = (Expr) stack.pop(Type.FLOAT);
        Expr left = (Expr) stack.pop(Type.FLOAT);
        Expr top = new ArithExpr(ArithExpr.MUL, left, right, Type.FLOAT);
        stack.push(top);
    }

    public void visitDMUL(DMUL inst) {
        Expr right = (Expr) stack.pop(Type.DOUBLE);
        Expr left = (Expr) stack.pop(Type.DOUBLE);
        Expr top = new ArithExpr(ArithExpr.MUL, left, right, Type.DOUBLE);
        stack.push(top);
    }

    public void visitIDIV(IDIV inst) {
        Expr right = (Expr) stack.pop(Type.INT);
        Expr left = (Expr) stack.pop(Type.INT);
        Expr check = new ZeroCheckExpr(right, Type.INT);
        Expr top = new ArithExpr(ArithExpr.DIV, left, check, Type.INT);
        stack.push(top);
    }

    public void visitLDIV(LDIV inst) {
        Expr right = (Expr) stack.pop(Type.LONG);
        Expr left = (Expr) stack.pop(Type.LONG);
        Expr check = new ZeroCheckExpr(right, Type.LONG);
        Expr top = new ArithExpr(ArithExpr.DIV, left, check, Type.LONG);
        stack.push(top);
    }

    public void visitFDIV(FDIV inst) {
        Expr right = (Expr) stack.pop(Type.FLOAT);
        Expr left = (Expr) stack.pop(Type.FLOAT);
        Expr top = new ArithExpr(ArithExpr.DIV, left, right, Type.FLOAT);
        stack.push(top);
    }

    public void visitDDIV(DDIV inst) {
        Expr right = (Expr) stack.pop(Type.DOUBLE);
        Expr left = (Expr) stack.pop(Type.DOUBLE);
        Expr top = new ArithExpr(ArithExpr.DIV, left, right, Type.DOUBLE);
        stack.push(top);
    }

    public void visitIREM(IREM inst) {
        Expr right = (Expr) stack.pop(Type.INT);
        Expr left = (Expr) stack.pop(Type.INT);
        Expr check = new ZeroCheckExpr(right, Type.INT);
        Expr top = new ArithExpr(ArithExpr.REM, left, check, Type.INT);
        stack.push(top);
    }

    public void visitLREM(LREM inst) {
        Expr right = (Expr) stack.pop(Type.LONG);
        Expr left = (Expr) stack.pop(Type.LONG);
        Expr check = new ZeroCheckExpr(right, Type.LONG);
        Expr top = new ArithExpr(ArithExpr.REM, left, check, Type.LONG);
        stack.push(top);
    }

    public void visitFREM(FREM inst) {
        Expr right = (Expr) stack.pop(Type.FLOAT);
        Expr left = (Expr) stack.pop(Type.FLOAT);
        Expr top = new ArithExpr(ArithExpr.REM, left, right, Type.FLOAT);
        stack.push(top);
    }

    public void visitDREM(DREM inst) {
        Expr right = (Expr) stack.pop(Type.DOUBLE);
        Expr left = (Expr) stack.pop(Type.DOUBLE);
        Expr top = new ArithExpr(ArithExpr.REM, left, right, Type.DOUBLE);
        stack.push(top);
    }

    /**
     * All <tt>visit_<i>x</i>neg</tt> push a NegExpr onto the stack.
     * 
     * @see NegExpr
     */
    public void visitINEG(INEG inst) {
        Expr expr = (Expr) stack.pop(Type.INT);
        Expr top = new NegExpr(expr, Type.INT);
        stack.push(top);
    }

    public void visitLNEG(LNEG inst) {
        Expr expr = (Expr) stack.pop(Type.LONG);
        Expr top = new NegExpr(expr, Type.LONG);
        stack.push(top);
    }

    public void visitFNEG(FNEG inst) {
        Expr expr = (Expr) stack.pop(Type.FLOAT);
        Expr top = new NegExpr(expr, Type.FLOAT);
        stack.push(top);
    }

    public void visitDNEG(DNEG inst) {
        Expr expr = (Expr) stack.pop(Type.DOUBLE);
        Expr top = new NegExpr(expr, Type.DOUBLE);
        stack.push(top);
    }

    /**
     * All <tt>visit_<i>x</i>sh<i>d</i></tt> push a ShiftExpr onto the
     * operand stack.
     * 
     * @see ShiftExpr
     */
    public void visitISHL(ISHL inst) {
        Expr right = (Expr) stack.pop(Type.INT);
        Expr left = (Expr) stack.pop(Type.INT);
        Expr top = new ShiftExpr(ShiftExpr.LEFT, left, right, Type.INT);
        stack.push(top);
    }

    public void visitLSHL(LSHL inst) {
        Expr right = (Expr) stack.pop(Type.INT);
        Expr left = (Expr) stack.pop(Type.LONG);
        Expr top = new ShiftExpr(ShiftExpr.LEFT, left, right, Type.LONG);
        stack.push(top);
    }

    public void visitISHR(ISHR inst) {
        Expr right = (Expr) stack.pop(Type.INT);
        Expr left = (Expr) stack.pop(Type.INT);
        Expr top = new ShiftExpr(ShiftExpr.RIGHT, left, right, Type.INT);
        stack.push(top);
    }

    public void visitLSHR(LSHR inst) {
        Expr right = (Expr) stack.pop(Type.INT);
        Expr left = (Expr) stack.pop(Type.LONG);
        Expr top = new ShiftExpr(ShiftExpr.RIGHT, left, right, Type.LONG);
        stack.push(top);
    }

    public void visitIUSHR(IUSHR inst) {
        Expr right = (Expr) stack.pop(Type.INT);
        Expr left = (Expr) stack.pop(Type.INT);
        Expr top = new ShiftExpr(ShiftExpr.UNSIGNED_RIGHT, left, right,
                Type.INT);
        stack.push(top);
    }

    public void visitLUSHR(LUSHR inst) {
        Expr right = (Expr) stack.pop(Type.INT);
        Expr left = (Expr) stack.pop(Type.LONG);
        Expr top = new ShiftExpr(ShiftExpr.UNSIGNED_RIGHT, left, right,
                Type.LONG);
        stack.push(top);
    }

    /**
     * All <tt>visit_<i>x op</i></tt> push an ArithExpr onto the stack.
     * 
     * @see ArithExpr
     */
    public void visitIAND(IAND inst) {
        Expr right = (Expr) stack.pop(Type.INT);
        Expr left = (Expr) stack.pop(Type.INT);
        Expr top = new ArithExpr(ArithExpr.AND, left, right, Type.INT);
        stack.push(top);
    }

    public void visitLAND(LAND inst) {
        Expr right = (Expr) stack.pop(Type.LONG);
        Expr left = (Expr) stack.pop(Type.LONG);
        Expr top = new ArithExpr(ArithExpr.AND, left, right, Type.LONG);
        stack.push(top);
    }

    public void visitIOR(IOR inst) {
        Expr right = (Expr) stack.pop(Type.INT);
        Expr left = (Expr) stack.pop(Type.INT);
        Expr top = new ArithExpr(ArithExpr.IOR, left, right, Type.INT);
        stack.push(top);
    }

    public void visitLOR(LOR inst) {
        Expr right = (Expr) stack.pop(Type.LONG);
        Expr left = (Expr) stack.pop(Type.LONG);
        Expr top = new ArithExpr(ArithExpr.IOR, left, right, Type.LONG);
        stack.push(top);
    }

    public void visitIXOR(IXOR inst) {
        Expr right = (Expr) stack.pop(Type.INT);
        Expr left = (Expr) stack.pop(Type.INT);
        Expr top = new ArithExpr(ArithExpr.XOR, left, right, Type.INT);
        stack.push(top);
    }

    public void visitLXOR(LXOR inst) {
        Expr right = (Expr) stack.pop(Type.LONG);
        Expr left = (Expr) stack.pop(Type.LONG);
        Expr top = new ArithExpr(ArithExpr.XOR, left, right, Type.LONG);
        stack.push(top);
    }

    /**
     * Visiting an iinc involves creating a ConstantExpr, LocalExpr, ArithExpr
     * StoreExpr, and a ExprStmt.
     */
    public void visitIINC(IINC iInc) {
        int incr = iInc.getIncrement();

        if (incr < 0) {
            Expr right = new ConstantExpr(new Integer(-incr), Type.INT);
            Expr left = new LocalExpr(iInc.getIndex(), Type.INT);

            Expr top = new ArithExpr(ArithExpr.SUB, left, right, Type.INT);

            LocalExpr copy = (LocalExpr) left.clone();
            copy.setDef(null);
            addStmt(new ExprStmt(new StoreExpr(copy, top, left.type())));
        } else if (incr > 0) {
            Expr right = new ConstantExpr(new Integer(incr), Type.INT);
            Expr left = new LocalExpr(iInc.getIndex(), Type.INT);

            Expr top = new ArithExpr(ArithExpr.ADD, left, right, Type.INT);

            LocalExpr copy = (LocalExpr) left.clone();
            copy.setDef(null);
            addStmt(new ExprStmt(new StoreExpr(copy, top, left.type())));
        }
    }

    /**
     * All cast visitors push a CastExpr onto the operand stack.
     */
    public void visitI2L(I2L inst) {
        Expr expr = (Expr) stack.pop(Type.INT);
        Expr top = new CastExpr(expr, Type.LONG, Type.LONG);
        stack.push(top);
    }

    public void visitI2F(I2F inst) {
        Expr expr = (Expr) stack.pop(Type.INT);
        Expr top = new CastExpr(expr, Type.FLOAT, Type.FLOAT);
        stack.push(top);
    }

    public void visitI2D(I2D inst) {
        Expr expr = (Expr) stack.pop(Type.INT);
        Expr top = new CastExpr(expr, Type.DOUBLE, Type.DOUBLE);
        stack.push(top);
    }

    public void visitL2I(L2I inst) {
        Expr expr = (Expr) stack.pop(Type.LONG);
        Expr top = new CastExpr(expr, Type.INT, Type.INT);
        stack.push(top);
    }

    public void visitL2F(L2F inst) {
        Expr expr = (Expr) stack.pop(Type.LONG);
        Expr top = new CastExpr(expr, Type.FLOAT, Type.FLOAT);
        stack.push(top);
    }

    public void visitL2D(L2D inst) {
        Expr expr = (Expr) stack.pop(Type.LONG);
        Expr top = new CastExpr(expr, Type.DOUBLE, Type.DOUBLE);
        stack.push(top);
    }

    public void visitF2I(F2I inst) {
        Expr expr = (Expr) stack.pop(Type.FLOAT);
        Expr top = new CastExpr(expr, Type.INT, Type.INT);
        stack.push(top);
    }

    public void visitF2L(F2L inst) {
        Expr expr = (Expr) stack.pop(Type.FLOAT);
        Expr top = new CastExpr(expr, Type.LONG, Type.LONG);
        stack.push(top);
    }

    public void visitF2D(F2D inst) {
        Expr expr = (Expr) stack.pop(Type.FLOAT);
        Expr top = new CastExpr(expr, Type.DOUBLE, Type.DOUBLE);
        stack.push(top);
    }

    public void visitD2I(D2I inst) {
        Expr expr = (Expr) stack.pop(Type.DOUBLE);
        Expr top = new CastExpr(expr, Type.INT, Type.INT);
        stack.push(top);
    }

    public void visitD2L(D2L inst) {
        Expr expr = (Expr) stack.pop(Type.DOUBLE);
        Expr top = new CastExpr(expr, Type.LONG, Type.LONG);
        stack.push(top);
    }

    public void visitD2F(D2F inst) {
        Expr expr = (Expr) stack.pop(Type.DOUBLE);
        Expr top = new CastExpr(expr, Type.FLOAT, Type.FLOAT);
        stack.push(top);
    }

    public void visitI2B(I2B inst) {
        Expr expr = (Expr) stack.pop(Type.INT);
        Expr top = new CastExpr(expr, Type.BYTE, Type.INT);
        stack.push(top);
    }

    public void visitI2C(I2C inst) {
        Expr expr = (Expr) stack.pop(Type.INT);
        Expr top = new CastExpr(expr, Type.CHAR, Type.INT);
        stack.push(top);
    }

    public void visitI2S(I2S inst) {
        Expr expr = (Expr) stack.pop(Type.INT);
        Expr top = new CastExpr(expr, Type.SHORT, Type.INT);
        stack.push(top);
    }

    /**
     * All <tt>visit_<i>x</i>cmp</tt> push an ArithExpr onto the stack.
     * 
     * @see ArithExpr
     */
    public void visitLCMP(LCMP inst) {
        Expr right = (Expr) stack.pop(Type.LONG);
        Expr left = (Expr) stack.pop(Type.LONG);
        Expr top = new ArithExpr(ArithExpr.CMP, left, right, Type.INT);
        stack.push(top);
    }

    public void visitFCMPL(FCMPL inst) {
        Expr right = (Expr) stack.pop(Type.FLOAT);
        Expr left = (Expr) stack.pop(Type.FLOAT);
        Expr top = new ArithExpr(ArithExpr.CMPL, left, right, Type.INT);
        stack.push(top);
    }

    public void visitFCMPG(FCMPG inst) {
        Expr right = (Expr) stack.pop(Type.FLOAT);
        Expr left = (Expr) stack.pop(Type.FLOAT);
        Expr top = new ArithExpr(ArithExpr.CMPG, left, right, Type.INT);
        stack.push(top);
    }

    public void visitDCMPL(DCMPL inst) {
        Expr right = (Expr) stack.pop(Type.DOUBLE);
        Expr left = (Expr) stack.pop(Type.DOUBLE);
        Expr top = new ArithExpr(ArithExpr.CMPL, left, right, Type.INT);
        stack.push(top);
    }

    public void visitDCMPG(DCMPG inst) {
        Expr right = (Expr) stack.pop(Type.DOUBLE);
        Expr left = (Expr) stack.pop(Type.DOUBLE);
        Expr top = new ArithExpr(ArithExpr.CMPG, left, right, Type.INT);
        stack.push(top);
    }

    /**
     * All <tt>visit_<i>x</i>eg</tt> add an IfZeroStmt to the statement
     * list.
     * 
     * @see IfZeroStmt
     */
    public void visitIFEQ(IFEQ bInst) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();

        Expr left = stack.pop(Type.INT);
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        // this could also maybe used getIndex??
        Assert.isTrue(t != null, "No block for " + bInst);

        addStmt(new IfZeroStmt(IfStmt.EQ, left, t, next));
    }

    public void visitIFNE(IFNE bInst) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();

        Expr left = stack.pop(Type.INT);
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        // this could also maybe used getIndex??
        Assert.isTrue(t != null, "No block for " + bInst);

        addStmt(new IfZeroStmt(IfStmt.NE, left, t, next));
    }

    public void visitIFLT(IFLT bInst) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();

        Expr left = stack.pop(Type.INT);
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        Assert.isTrue(t != null, "No block for " + bInst);

        addStmt(new IfZeroStmt(IfStmt.LT, left, t, next));
    }

    public void visitIFGE(IFGE bInst) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();

        Expr left = stack.pop(Type.INT);
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        Assert.isTrue(t != null, "No block for " + bInst);

        addStmt(new IfZeroStmt(IfStmt.GE, left, t, next));
    }

    public void visitIFGT(IFGT bInst) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();

        Expr left = stack.pop(Type.INT);
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        Assert.isTrue(t != null, "No block for " + bInst);

        addStmt(new IfZeroStmt(IfStmt.GT, left, t, next));
    }

    public void visitIFLE(IFLE bInst) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();

        Expr left = stack.pop(Type.INT);
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        Assert.isTrue(t != null, "No block for " + bInst);

        addStmt(new IfZeroStmt(IfStmt.LE, left, t, next));
    }

    /**
     * All <tt>visit_if_<i>x</i>cmp<i>y</i></tt> add a IfCmpStmt to the
     * statement list.
     */
    public void visitIF_ICMPEQ(IF_ICMPEQ bInst) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();

        Expr right = stack.pop(Type.INT);
        Expr left = stack.pop(Type.INT);
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        Assert.isTrue(t != null, "No block for " + bInst);

        addStmt(new IfCmpStmt(IfStmt.EQ, left, right, t, next));
    }

    public void visitIF_ICMPNE(IF_ICMPNE bInst) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();

        Expr right = stack.pop(Type.INT);
        Expr left = stack.pop(Type.INT);
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        Assert.isTrue(t != null, "No block for " + bInst);

        addStmt(new IfCmpStmt(IfStmt.NE, left, right, t, next));
    }

    public void visitIF_ICMPLT(IF_ICMPLT bInst) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();

        Expr right = stack.pop(Type.INT);
        Expr left = stack.pop(Type.INT);
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        Assert.isTrue(t != null, "No block for " + bInst);

        addStmt(new IfCmpStmt(IfStmt.LT, left, right, t, next));
    }

    public void visitIF_ICMPGE(IF_ICMPGE bInst) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();

        Expr right = stack.pop(Type.INT);
        Expr left = stack.pop(Type.INT);
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        Assert.isTrue(t != null, "No block for " + bInst);

        addStmt(new IfCmpStmt(IfStmt.GE, left, right, t, next));
    }

    public void visitIF_ICMPGT(IF_ICMPGT bInst) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();

        Expr right = stack.pop(Type.INT);
        Expr left = stack.pop(Type.INT);
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        Assert.isTrue(t != null, "No block for " + bInst);

        addStmt(new IfCmpStmt(IfStmt.GT, left, right, t, next));
    }

    public void visitIF_ICMPLE(IF_ICMPLE bInst) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();

        Expr right = stack.pop(Type.INT);
        Expr left = stack.pop(Type.INT);
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        Assert.isTrue(t != null, "No block for " + bInst);

        addStmt(new IfCmpStmt(IfStmt.LE, left, right, t, next));
    }

    public void visitIF_ACMPEQ(IF_ACMPEQ bInst) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();

        Expr right = stack.pop(Type.OBJECT);
        Expr left = stack.pop(Type.OBJECT);
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        Assert.isTrue(t != null, "No block for " + bInst);

        addStmt(new IfCmpStmt(IfStmt.EQ, left, right, t, next));
    }

    public void visitIF_ACMPNE(IF_ACMPNE bInst) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();

        Expr right = stack.pop(Type.OBJECT);
        Expr left = stack.pop(Type.OBJECT);
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        Assert.isTrue(t != null, "No block for " + bInst);

        addStmt(new IfCmpStmt(IfStmt.NE, left, right, t, next));
    }

    /**
     * Adds a GotoStmt to the statement list.
     * 
     * @see GotoStmt
     */
    public void visitGOTO(GOTO bInst) {
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        Assert.isTrue(t != null, "No block for " + bInst);
        addStmt(new GotoStmt(t));
    }

    public void visitGOTO_W(GOTO_W bInst) {
        Block t = (Block) block.graph().getNode(bInst.getTarget());
        Assert.isTrue(t != null, "No block for " + bInst);
        addStmt(new GotoStmt(t));
    }

    /**
     * Adds a JsrStmt to the statement list.
     * 
     * @see JsrStmt
     */
    public void visitJsrInstruction(JsrInstruction bInst) {
        // Push the return address after we add the statement.
        // This prevents it from being saved to a local variable.
        // It's illegal to load a return address from a local variable,
        // so we can't save it.
        Subroutine sub = block.graph().labelSub(bInst.getTarget());
        addStmt(new JsrStmt(sub, next));
        stack.push(new ReturnAddressExpr(ReturnaddressType.NO_TARGET));
        // AGAIN THIS MAY NOT BE THE BEST
    }

    /**
     * Adds a RetStmt to the statement list.
     * 
     * @see RetStmt
     */
    public void visitRET(RET inst) {
        Assert.isTrue(sub != null);
        addStmt(new RetStmt(sub));
    }

    /**
     * Add a SwitchStmt to the statement list.
     * 
     * @see SwitchStmt
     */
    public void visitSelect(Select sInst) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();
        Expr index = stack.pop(Type.INT);

        Block defaultTarget = (Block) block.graph().getNode(sInst.getTarget());
        Assert.isTrue(defaultTarget != null, "No block for " + sInst);

        InstructionHandle[] targ = sInst.getTargets();
        Block[] targets = new Block[targ.length];

        for (int i = 0; i < targets.length; i++) {
            targets[i] = (Block) block.graph().getNode(targ[i]);
            Assert.isTrue(targets[i] != null, "No block for " + sInst);
        } // LABEL STUFF DONE HERE

        addStmt(new SwitchStmt(index, defaultTarget, targets, sInst.getMatchs()));
    }

    /**
     * All <tt>visit_<i>x</i>return</tt> add a ReturnExprStmt to the
     * statement list.
     * 
     * @see ReturnExprStmt
     */
    public void visitIRETURN(IRETURN inst) {
        Expr expr = (Expr) stack.pop(Type.INT);
        addStmt(new ReturnExprStmt(expr));
    }

    public void visitLRETURN(LRETURN inst) {
        Expr expr = (Expr) stack.pop(Type.LONG);
        addStmt(new ReturnExprStmt(expr));
    }

    public void visitFRETURN(FRETURN inst) {
        Expr expr = (Expr) stack.pop(Type.FLOAT);
        addStmt(new ReturnExprStmt(expr));
    }

    public void visitDRETURN(DRETURN inst) {
        Expr expr = (Expr) stack.pop(Type.DOUBLE);
        addStmt(new ReturnExprStmt(expr));
    }

    public void visitARETURN(ARETURN inst) {
        Expr expr = (Expr) stack.pop(Type.OBJECT);
        addStmt(new ReturnExprStmt(expr));
    }

    /**
     * Adds a ReturnStmt to the statement list.
     */
    public void visitRETURN(RETURN inst) {
        addStmt(new ReturnStmt());
    }

    /**
     * Pushes a StaticFieldExpr onto the operand stack.
     */
    public void visitGETSTATIC(GETSTATIC statInst) {
        ConstantPoolGen cp = block.graph().method().getConstantPool();
        MemberRef field = new MemberRef((ObjectType) statInst
                .getReferenceType(cp), new NameAndType(statInst
                .getFieldName(cp), statInst.getFieldType(cp)));
        Type type = statInst.getFieldType(cp);
        try {
            JavaClass jc = org.apache.bcel.Repository.lookupClass(statInst
                    .getLoadClassType(cp).getClassName());
            if (Tree.DEBUG && jc == null)
                System.out.println("ARGGGH: no class found "
                        + statInst.getLoadClassType(cp).getClassName());
            Field[] fields = jc.getFields();
            int j = -1;
            for (int i = 0; i < fields.length; i++) {
                if (fields[i].getName().equals(statInst.getFieldName(cp))) {
                    j = i;
                    break;
                }
            }
            if (j == -1) {
                throw new NoSuchFieldException();
            }
            Field e = fields[j];

            if (e.isFinal()) {
                if (e.getConstantValue() != null) {// WHY??MAYBE DODGY?
                    // There doesn't seem to be much way to check the
                    // constant's value in BCEL
                    // and BLOAT's is an Object so could be anything...
                    int index = (e.getConstantValue()).getConstantValueIndex();
                    Constant c = cp.getConstant(index);
                    if ((c instanceof ConstantObject)
                            && (c instanceof ConstantClass)) {
                        ConstantObject co = (ConstantObject) c;
                        Expr top = new ConstantExpr(co.getConstantValue(cp
                                .getConstantPool()), type);
                        stack.push(top);
                        // context.release(e.fieldInfo());//We never started
                        // editing it
                        return;
                    }
                }
            }

            // context.release(e.fieldInfo());//We never started editing it
        } catch (NoSuchFieldException e) {
            // No field found. Silently assume non-final.
        } catch (ClassNotFoundException e) {
            // Ignore for the moment.
        }

        Expr top = new StaticFieldExpr(field, type);
        stack.push(top);
    }

    public void visitPUTSTATIC(PUTSTATIC pS) {
        ConstantPoolGen cp = block.graph().method().getConstantPool();
        MemberRef field = new MemberRef((ObjectType) pS.getReferenceType(cp),
                new NameAndType(pS.getFieldName(cp), pS.getFieldType(cp)));
        Type type = field.nameAndType().type();
        Expr value = stack.pop(type);
        StaticFieldExpr target = new StaticFieldExpr(field, type);
        addStore(target, value);
    }

    // public void visit_putstatic_nowb( inst) {
    // visit_putstatic(inst);
    // }

    /**
     * Pushes a FieldExpr onto the operand stack.
     */
    public void visitGETFIELD(GETFIELD gF) {
        ConstantPoolGen cp = block.graph().method().getConstantPool();
        MemberRef field = new MemberRef((ObjectType) gF.getReferenceType(cp),
                new NameAndType(gF.getFieldName(cp), gF.getFieldType(cp)));
        Type type = field.nameAndType().type();
        Expr obj = stack.pop(Type.OBJECT);
        Expr check = new ZeroCheckExpr(obj, obj.type());
        Expr top = new FieldExpr(check, field, type);
        stack.push(top);
    }

    public void visitPUTFIELD(PUTFIELD pF) {
        ConstantPoolGen cp = block.graph().method().getConstantPool();
        MemberRef field = new MemberRef((ObjectType) pF.getReferenceType(cp),
                new NameAndType(pF.getFieldName(cp), pF.getFieldType(cp)));
        Type type = field.nameAndType().type();
        Expr value = stack.pop(type);
        Expr obj = stack.pop(Type.OBJECT);
        Expr ucCheck = obj;

        if (USE_PERSISTENT)
            ucCheck = new UCExpr(obj, UCExpr.POINTER, obj.type());

        Expr check = new ZeroCheckExpr(ucCheck, obj.type());

        new FieldExpr(check, field, type);

        FieldExpr target = new FieldExpr(check, field, type);
        addStore(target, value);
    }

    // Don't insert UCExpr
    /*
     * Persistentcy Stuff public void visit_putfield_nowb(Instruction inst) {
     * PUTFIELD pF = (PUTFIELD)inst; ConstantPoolGen cp =
     * block.graph().method().getConstantPool(); MemberRef field = new
     * MemberRef(pF.getClassType(cp), new NameAndType(pF.getFieldName(cp),
     * pF.getFieldType(cp))); Type type = field.nameAndType().type(); Expr value =
     * stack.pop(type); Expr obj = stack.pop(Type.OBJECT); Expr check = new
     * ZeroCheckExpr(obj, obj.type()); Expr top = new FieldExpr(check, field,
     * type); FieldExpr target = new FieldExpr(check, field, type);
     * addStore(target, value); }
     */

    /**
     * All <tt>visit_invoke<i>x</i></tt> deal with a CallMethodExpr or a
     * CallStaticExpr.
     * 
     * @see CallMethodExpr
     * @see CallStaticExpr
     */
    public void visitINVOKEVIRTUAL(INVOKEVIRTUAL inst) {
        addCall(inst, CallMethodExpr.VIRTUAL);
    }

    public void visitINVOKESPECIAL(INVOKESPECIAL inst) {
        addCall(inst, CallMethodExpr.NONVIRTUAL);
    }

    public void visitINVOKESTATIC(INVOKESTATIC inst) {
        addCall(inst, 0);
    }

    public void visitINVOKEINTERFACE(INVOKEINTERFACE inst) {
        addCall(inst, CallMethodExpr.INTERFACE);
    }

    /**
     * Creates either a CallMethodExpr or a CallStaticExpr to represent a method
     * call. After obtaining some information about the method. The parameters
     * to the methods are popped from the operand stack. If the method is not
     * static, the "this" object is popped from the operand stack. A
     * CallMethodExpr is created for a non-static method and a CallStaticExpr is
     * created for a static method. If the method returns a value, the Call*Expr
     * is pushed onto the stack. If the method has no return value, it is
     * wrapped in an ExprStmt and is added to the statement list.
     */
    private void addCall(InvokeInstruction iI, int kind) {
        ConstantPoolGen cp = block.graph().method().getConstantPool();
        Type[] paramTypes = iI.getArgumentTypes(cp);
        Type returnType = iI.getReturnType(cp);
        ReferenceType caller = iI.getReferenceType(cp);

        MethodRef method = new MethodRef(caller, new NameAndType(iI
                .getMethodName(cp), returnType), returnType, paramTypes);

        Expr[] params = new Expr[paramTypes.length];

        for (int i = paramTypes.length - 1; i >= 0; i--) {
            params[i] = stack.pop(paramTypes[i]);
        }

        Expr top;

        if (!(iI instanceof INVOKESTATIC)) {
            Expr obj = stack.pop(Type.OBJECT);

            top = new CallMethodExpr(kind, obj, params, method);

        } else {
            top = new CallStaticExpr(params, method);
        }

        if (returnType.equals(Type.VOID)) {
            addStmt(new ExprStmt(top));

        } else {
            stack.push(top);
        }
    }

    /**
     * Pushes a NewExpr onto the operand stack.
     * 
     * @see NewExpr
     */
    public void visitNEW(NEW nI) {
        ConstantPoolGen cp = block.graph().method().getConstantPool();
        Type type = nI.getLoadClassType(cp);
        Expr top = new NewExpr(type, Type.OBJECT);
        stack.push(top);

        db("      new: " + top);
    }

    /**
     * Pushes a NewArrayExpr onto the operand stack.
     */
    public void visitNEWARRAY(NEWARRAY nAI) {
        BasicType type = (BasicType) ((ArrayType) nAI.getType()).getBasicType();
        Expr size = stack.pop(Type.INT);
        Expr top = new NewArrayExpr(size, type, (ArrayType) nAI.getType());
        stack.push(top);
    }

    /**
     * Pushes a NewArrayExpr onto the operand stack.
     */
    public void visitANEWARRAY(ANEWARRAY nAI) {
        ConstantPoolGen cp = block.graph().method().getConstantPool();
        Expr size = stack.pop(Type.INT);
        Expr top = new NewArrayExpr(size, nAI.getType(cp), new ArrayType(nAI
                .getType(cp), 1));
        stack.push(top);
    }

    /**
     * Pushes an ArrayLengthExpr onto the operand stack.
     * 
     * @see ArrayLengthExpr
     */
    public void visitARRAYLENGTH(ARRAYLENGTH inst) {
        Expr array = (Expr) stack.pop(Type.OBJECT);
        Expr top = new ArrayLengthExpr(array, Type.INT);
        stack.push(top);
    }

    /**
     * Adds a ThrowStmt to the statement list.
     * 
     * @see ThrowStmt
     */
    public void visitATHROW(ATHROW inst) {
        Expr expr = (Expr) stack.pop(ObjectType.THROWABLE);
        addStmt(new ThrowStmt(expr));
    }

    /**
     * Pushes a CastExpr onto the operand stack.
     * 
     * @see CastExpr
     */
    public void visitCHECKCAST(CHECKCAST cCI) {
        Expr expr = stack.pop(Type.OBJECT);
        ConstantPoolGen cp = block.graph().method().getConstantPool();
        Type type = cCI.getType(cp);
        Expr top = new CastExpr(expr, type, type);
        stack.push(top);
    }

    /**
     * Pushes an InstanceOfExpr onto the operand stack.
     * 
     * @see InstanceOfExpr
     */
    public void visitINSTANCEOF(INSTANCEOF iOI) {
        ConstantPoolGen cp = block.graph().method().getConstantPool();
        Type type = iOI.getType(cp);
        Expr expr = (Expr) stack.pop(Type.OBJECT);
        Expr top = new InstanceOfExpr(expr, type, Type.INT);
        stack.push(top);
    }

    /**
     * Both <tt>monitor</tt> visitors add a MonitorStmt to the statement list.
     * 
     * @see MonitorStmt
     */
    public void visitMONITORENTER(MONITORENTER inst) {
        Expr obj = (Expr) stack.pop(Type.OBJECT);
        addStmt(new MonitorStmt(MonitorStmt.ENTER, obj));
    }

    public void visitMONITOREXIT(MONITOREXIT inst) {
        Expr obj = (Expr) stack.pop(Type.OBJECT);
        addStmt(new MonitorStmt(MonitorStmt.EXIT, obj));
    }

    /**
     * Push a NewMultiArrayExpr onto the operand stack.
     * 
     * @see NewMultiArrayExpr
     */
    public void visitMULTIANEWARRAY(MULTIANEWARRAY mANAI) {
        ConstantPoolGen cp = block.graph().method().getConstantPool();

        Expr[] dim = new Expr[mANAI.getDimensions()];

        for (int i = dim.length - 1; i >= 0; i--) {
            dim[i] = stack.pop(Type.INT);
        }

        Type type = mANAI.getType(cp);
        Expr top = new NewMultiArrayExpr(dim, type, new ArrayType(type,
                dim.length));

        stack.push(top);
    }

    /**
     * Both <tt>visit_<i>x</i>null</tt> add an IfZeroStmt to the statement
     * list.
     * 
     * ssee IfZeroStmt
     */
    public void visitIFNULL(IFNULL bI) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();
        Expr left = stack.pop(Type.OBJECT);

        Block t = (Block) block.graph().getNode(bI.getTarget());
        Assert.isTrue(t != null, "No block for " + bI);

        addStmt(new IfZeroStmt(IfStmt.EQ, left, t, next));
    }

    public void visitIFNONNULL(IFNONNULL bI) {
        // It isn't necessary to saveStack since removing critical edges
        // guarantees that no code will be inserted before the branch
        // since there cannot be more than one predecessor of either of
        // the successor nodes (and hence no phi statements, even in SSAPRE).
        // saveStack();

        Expr left = stack.pop(Type.OBJECT);
        Block t = (Block) block.graph().getNode(bI.getTarget());
        Assert.isTrue(t != null, "No block for " + bI);

        addStmt(new IfZeroStmt(IfStmt.NE, left, t, next));
    }

    /**
     * Replaces the expression on the top of the stack with an RCExpr.
     * 
     * @see RCExpr
     */
    // Persistency Stuff?
    /*
     * public void visit_rc(Instruction inst) { Integer depth = (Integer)
     * inst.operand(); Expr object = stack.peek(depth.intValue());
     * stack.replace(depth.intValue(), new RCExpr(object, object.type())); }
     */
    /**
     * 
     */
    // More Persistency Stuff
    /*
     * public void visit_aupdate(Instruction inst) { Integer depth = (Integer)
     * inst.operand();
     * 
     * if (AUPDATE_FIX_HACK) { // Hack to fix a bug in old bloat-generated code:
     * 
     * if (depth.intValue() == 1) { Expr object = stack.peek();
     * 
     * if (object.type().isWide()) { depth = new Integer(2);
     * inst.setOperand(depth); AUPDATE_FIX_HACK_CHANGED = true; } } }
     * 
     * Expr object = stack.peek(depth.intValue());
     * stack.replace(depth.intValue(), new UCExpr(object, UCExpr.POINTER,
     * object.type())); }
     */
    /**
     * Replace the expression at the stack depth specified in the instruction
     * with a UCExpr.
     * 
     * @see UCExpr
     */
    // More Persistency Stuff
    /*
     * public void visit_supdate(Instruction inst) { Integer depth = (Integer)
     * inst.operand(); Expr object = stack.peek(depth.intValue());
     * stack.replace(depth.intValue(), new UCExpr(object, UCExpr.SCALAR,
     * object.type())); }
     */
    /**
     * Add a SCStmt to the statement list
     * 
     * @see SCStmt
     */
    // Tempory we might have to Create a ASWIZZLE instruction?
    /*
     * public void visit_aswizzle(ASWIZZLE inst) { Expr index = (Expr)
     * stack.pop(Type.INT); Expr array = (Expr) stack.pop(new
     * ArrayType(Type.OBJECT,1));
     * 
     * addStmt(new SCStmt(array, index)); }
     */
    /**
     * Add a SRStmt to the statement list.
     * 
     * @see SRStmt
     */
    /*
     * public void visit_aswrange(Instruction inst) { Expr end = (Expr)
     * stack.pop(Type.INT); Expr start = (Expr) stack.pop(Type.INT); Expr array =
     * (Expr) stack.pop(new ArrayType(Type.OBJECT,1));
     * 
     * addStmt(new SRStmt(array, start, end)); }
     */
    /**
     * Visit all the statements in the statement list.
     */
    public void visitForceChildren(TreeVisitor visitor) {
        LinkedList list = new LinkedList(stmts);

        if (visitor.reverse()) {
            ListIterator iter = list.listIterator(stmts.size());

            while (iter.hasPrevious()) {
                Stmt s = (Stmt) iter.previous();
                s.visit(visitor);
            }
        } else {
            ListIterator iter = list.listIterator();

            while (iter.hasNext()) {
                Stmt s = (Stmt) iter.next();
                s.visit(visitor);
            }
        }
    }

    public void visit(TreeVisitor visitor) {
        visitor.visitTree(this);
    }

    public Node parent() {
        return null;
    }

    public Block block() {
        return block;
    }

    public void visitBREAKPOINT(BREAKPOINT inst) {
    }

    public void visitLOOKUPSWITCH(LOOKUPSWITCH inst) {
    }

    public void visitIMPDEP1(IMPDEP1 inst) {
    }

    public void visitIMPDEP2(IMPDEP2 inst) {
    }

    public void visitTABLESWITCH(TABLESWITCH inst) {
    }

    public void visitStackConsumer(StackConsumer inst) {
    }

    public void visitStackProducer(StackProducer inst) {
    }

    public void visitVariableLengthInstruction(VariableLengthInstruction inst) {
    }

    public void visitLoadInstruction(LoadInstruction inst) {
    }

    public void visitExceptionThrower(ExceptionThrower inst) {
    }

    public void visitConstantPushInstruction(ConstantPushInstruction inst) {
    }

    public void visitFieldOrMethod(org.apache.bcel.generic.FieldOrMethod inst) {
    }

    public void visitReturnInstruction(ReturnInstruction inst) {
    }

    public void visitAllocationInstruction(AllocationInstruction inst) {
    }

    public void visitArrayInstruction(ArrayInstruction inst) {
    }

    public void visitInvokeInstruction(InvokeInstruction inst) {
    }

    public void visitCPInstruction(CPInstruction inst) {
    }

    public void visitArithmeticInstruction(ArithmeticInstruction inst) {
    }

    public void visitPushInstruction(PushInstruction inst) {
    }

    public void visitUnconditionalBranch(UnconditionalBranch inst) {
    }

    public void visitGotoInstruction(GotoInstruction inst) {
    }

    public void visitTypedInstruction(TypedInstruction inst) {
    }

    public void visitStoreInstruction(StoreInstruction inst) {
    }

    public void visitPopInstruction(PopInstruction inst) {
    }

    public void visitConversionInstruction(ConversionInstruction inst) {
    }

    public void visitIfInstruction(IfInstruction inst) {
    }

    public void visitFieldInstruction(FieldInstruction inst) {
    }

    public void visitLoadClass(LoadClass inst) {
    }

    public void visitBranchInstruction(BranchInstruction inst) {
    }

    public void visitLocalVariableInstruction(LocalVariableInstruction inst) {
    }

    public void visitStackInstruction(StackInstruction inst) {
    }

    public void visitJSR(JSR inst) {
    }

    public void visitJSR_W(JSR_W inst) {
    }

}
